from ipywidgets.widgets import Widget
import ipyvuetify as v


def get_widget(root_widget: Widget, address: str | list[int]) -> Widget:
    """
    Returns a sub widget of root_widget. Address is a sequence of childhood ranks as a string
    Return sub_widget may be modified, it's still the same sub_widget of the root_widget
    get_widget(root_widget, '0') returns root_widgetn first child
    TODO : allow childhood rank > 9
    """
    if not address:
        return root_widget

    try:
        address = [int(i) for i in address]
    except ValueError:
        raise ValueError(address, "must be a string composed of digits")
    try:
        return recursive_get_widget(root_widget, address)
    except:
        raise IndexError(f"Nothing found @{address} in this {root_widget.__class__.__name__}")


def recursive_get_widget(root_widget: Widget, address: list[int]):
    if not address:
        return root_widget

    if isinstance(root_widget, v.Tooltip):
        assert address.pop(0) == 0
        new_root = root_widget.v_slots[0]["children"]
    else:
        new_root = root_widget.children[address.pop(0)]
    return recursive_get_widget(new_root, address)


def change_widget(root_widget: Widget, address: str, new_widget):
    """
    Substitutes a sub_widget in a root_widget.
    Address is a sequence of childhood ranks as a string, root_widget first child address is  '0'
    The root_widget is altered but the object remains the same
    """

    if not address:
        raise ValueError('must provide at least the parent widget')

    parent_widget = get_widget(root_widget, address[:-1])
    new_children = list(parent_widget.children)
    new_children[int(address[-1])] = new_widget
    parent_widget.children = new_children



import ipyvuetify as v
import pandas as pd
from antakia_core.data_handler.region import ModelRegion
from antakia_core.utils.utils import ProblemCategory

from antakia import config
from antakia.gui.graphical_elements.sub_model_table import SubModelTable
from antakia.gui.helpers.progress_bar import ProgressBar
from antakia.gui.tabs.model_explorer import ModelExplorer
from antakia.utils.stats import log_errors, stats_logger


class Tab3:
    headers = [
        {
            "text": column,
            "sortable": True,
            "value": column,
            # "class": "primary white--text",\
        }
        for column in ['Sub-model', 'MSE', 'MAE', 'R2', 'delta']
    ]

    def __init__(self, X: pd.DataFrame, problem_category: ProblemCategory, validate_callback: callable):
        self.X = X
        self.problem_category = problem_category
        self.validate_callback = validate_callback
        self.model_explorer = ModelExplorer(self.X)
        self.region: ModelRegion | None = None
        self.substitution_model_training = False  # tab 3 : training flag

        self._build_widget()
        self.progress_bar = ProgressBar(
            self.progress_wgt, indeterminate=True, reset_at_end=True
        )
        self.progress_bar(100)

    def _build_widget(self):
        """
        build the tab3 widget - part of init method
        Returns
        -------

        """
        self.validate_model_btn = v.Btn(  # 4501000
            v_on='tooltip.on',
            class_="ma-1 mt-12 green white--text",
            children=[
                v.Icon(
                    class_="mr-2",
                    children=[
                        "mdi-check"
                    ],
                ),
                "Validate sub-model",
            ],
        )
        self.model_table = SubModelTable(  # 45001
            headers=self.headers,
            items=[],
        )
        self.region_prefix_wgt = v.Html(class_="mr-2", tag="h3", children=["Region"])  # 450000
        self.region_chip_wgt = v.Chip(color="red", children=["1"], )  # 450001
        self.region_title = v.Html(class_="ml-2", tag="h3", children=[""])  # 450002
        self.progress_wgt = v.ProgressLinear(  # 450110
            style_="width: 100%",
            class_="mt-4",
            v_model=0,
            height="15",
            indeterminate=True,
            color="blue",
        )
        self.widget = [
            v.Row(  # 450
                class_="d-flex",
                children=[
                    v.Col(  # Col1 - sub model table #4500
                        class_="col-5",
                        children=[
                            v.Sheet(  # 45000
                                class_="ma-1 d-flex flex-row align-center",
                                children=[
                                    self.region_prefix_wgt,
                                    self.region_chip_wgt,
                                    self.region_title
                                ]
                            ),
                            self.model_table
                        ]
                    ),
                    v.Col(  # Col2 - buttons #4501
                        class_="col-2",
                        children=[
                            v.Row(
                                class_="flex-column",
                                children=[
                                    v.Tooltip(  # 45010
                                        bottom=True,
                                        v_slots=[
                                            {
                                                'name': 'activator',
                                                'variable': 'tooltip',
                                                'children':
                                                    self.validate_model_btn,
                                            }
                                        ],
                                        children=['Chose this submodel']
                                    )
                                ]
                            ),
                            v.Row(
                                class_="flex-column",
                                children=[
                                    self.progress_wgt
                                ]
                            )
                        ]
                    ),
                    v.Col(  # Col3 - model explorer #4502
                        class_="col-5",
                        children=[
                            self.model_explorer.widget
                        ]
                    ),
                ]
            )
        ]
        # We wire a select event on the 'substitution table' :
        self.model_table.set_callback(self._sub_model_selected_callback)

        # We wire a ckick event on the "validate sub-model" button :
        self.validate_model_btn.on_event("click", self._validate_sub_model)
        self.update()

    @property
    def selected_sub_model(self):
        return self.model_table.selected

    @selected_sub_model.setter
    def selected_sub_model(self, value):
        self.model_table.selected = value

    def update_region(self, region: ModelRegion, train=True):
        """
        method to update the region of substitution
        Parameters
        ----------
        region: region to substitute

        Returns
        -------

        """
        self.region = region
        if self.region is not None and train:
            # We update the substitution table once to show the name of the region
            self.substitution_model_training = True
            self.progress_bar(0)
            self.update()
            # show tab 3 (and update)
            self.region.train_substitution_models(task_type=self.problem_category)

            self.progress_bar(100)
            self.substitution_model_training = False
            self.update()
        else:
            self.update()

    def update(self):
        self._update_substitution_prefix()
        self._update_substitution_title()
        self._update_model_table()
        self._update_selected()
        self._update_validate_btn()

    def _update_substitution_prefix(self):
        # Region prefix text
        self.region_prefix_wgt.class_ = "mr-2 black--text" if self.region else "mr-2 grey--text"
        # v.Chip
        self.region_chip_wgt.color = self.region.color if self.region else "grey"
        self.region_chip_wgt.children = [str(self.region.num)] if self.region else ["-"]

    def _update_model_table(self):
        if (
            self.substitution_model_training or
            not self.region or
            self.region.num_points() < config.ATK_MIN_POINTS_NUMBER or
            len(self.region.perfs) == 0
        ):
            self.model_table.items = []
        else:
            def series_to_str(series: pd.Series) -> str:
                return series.apply(lambda x: f"{x:.2f}")

            perfs = self.region.perfs.copy()
            stats_logger.log('substitute_model', {'best_perf': perfs['delta'].min()})
            for col in perfs.columns:
                if col != 'delta_color':
                    perfs[col] = series_to_str(perfs[col])
            perfs = perfs.reset_index().rename(columns={"index": "Sub-model"})
            headers = [
                {
                    "text": column,
                    "sortable": False,
                    "value": column,
                }
                for column in perfs.drop('delta_color', axis=1).columns
            ]
            self.model_table.headers = headers
            self.model_table.items = perfs.to_dict("records")

    def _update_selected(self):
        if self.region and self.region.interpretable_models.selected_model:
            # we set to selected model if any
            self.model_table.selected = [{'Sub-model': self.region.interpretable_models.selected_model}]
            self.model_explorer.update_selected_model(self.region.get_selected_model(), self.region)
        else:
            # clear selection if new region:
            self.model_explorer.reset()
            self.model_table.selected = []

    def _update_substitution_title(self):
        title = self.region_title
        title.tag = "h3"
        table = self.model_table  # subModel table
        if self.substitution_model_training:
            # We tell to wait ...
            title.class_ = "ml-2 grey--text italic "
            title.children = [f"Sub-models are being evaluated ..."]
            # We clear items int the SubModelTable
        elif not self.region:  # no region provided
            title.class_ = "ml-2 grey--text italic "
            title.children = [f"No region selected for substitution"]
        elif self.region.num_points() < config.ATK_MIN_POINTS_NUMBER:  # region is too small
            title.class_ = "ml-2 red--text"
            title.children = ["Region too small for substitution !"]
        elif len(self.region.perfs) == 0:  # model not trained
            title.class_ = "ml-2 red--text"
            title.children = ["click on substitute button to train substitution models"]
        else:
            # We have results
            title.class_ = "ml-2 black--text"
            title.children = [
                f"{self.region.name}, "
                f"{self.region.num_points()} points, {100 * self.region.dataset_cov():.1f}% of the dataset"
            ]

    def _update_validate_btn(self):
        self.validate_model_btn.disabled = len(self.selected_sub_model) == 0

    def _sub_model_selected_callback(self, data):
        """
        callback on model selection - updates the model explorer
        Parameters
        ----------
        data

        Returns
        -------

        """
        is_selected = bool(data["value"])
        # We use this GUI attribute to store the selected sub-model
        self.selected_sub_model = [data['item']]
        self.validate_model_btn.disabled = not is_selected
        if is_selected:
            self.model_explorer.update_selected_model(self.region.get_model(data['item']['Sub-model']), self.region)
        else:
            self.model_explorer.reset()

    @log_errors
    def _validate_sub_model(self, *args):
        """
        callback called on model validation
        Parameters
        ----------
        args

        Returns
        -------

        """
        # We get the sub-model data from the SubModelTable:
        # get_widget(self.widget,"45001").items[self.validated_sub_model]

        self.validate_model_btn.disabled = True

        stats_logger.log('validate_sub_model', {'model': self.selected_sub_model[0]['Sub-model']})

        # We udpate the region
        self.region.select_model(self.selected_sub_model[0]['Sub-model'])
        self.region.validate()
        # empty selected region
        self.region = None
        self.selected_sub_model = []
        # Show tab 2
        self.validate_callback()



from typing import Callable

from functools import partial

import numpy as np
import pandas as pd

from sklearn.datasets import load_breast_cancer, make_blobs
from scipy.stats import multivariate_normal


def generate_corner_dataset(
        num_samples: int, corner_position: str = "top_right", random_seed: int | None = None
) -> (np.ndarray, np.ndarray):
    """Generate a toy dataset with a corner of the feature space.

    Parameters
    ----------
    num_samples : int
        Number of samples to generate.
    corner_position : str, optional
        Position of the corner in the feature space, by default "top_right"

    Returns
    -------
    (np.ndarray, np.ndarray)
        Tuple of feature matrix and target vector.

    Raises
    ------
    ValueError
        If corner_position is not one of "top_left", "top_right", "bottom_left", "bottom_right".
    """
    np.random.seed(random_seed)
    X = np.random.uniform(0, 1, (num_samples, 2))

    if corner_position == "top_right":
        mask = (X[:, 0] > 0.5) & (X[:, 1] > 0.5)
    elif corner_position == "top_left":
        mask = (X[:, 0] < 0.5) & (X[:, 1] > 0.5)
    elif corner_position == "bottom_right":
        mask = (X[:, 0] > 0.5) & (X[:, 1] < 0.5)
    elif corner_position == "bottom_left":
        mask = (X[:, 0] < 0.5) & (X[:, 1] < 0.5)

    else:
        raise ValueError(
            "Invalid corner position must be one of: top_right, top_left, bottom_right, bottom_left."
        )

    y = mask.astype(int)
    return X, y


def get_data_from_mixture_distribution(
        num_samples, positive_component_rvs: Callable, negative_component_rvs: Callable
):
    d = positive_component_rvs(size=1).shape[0]
    Y = np.random.binomial(1, 0.5, num_samples)
    X = np.zeros((num_samples, d))
    for i in range(num_samples):
        if Y[i] == 1:
            X[i] = positive_component_rvs(size=1)
        else:
            X[i] = negative_component_rvs(size=1)
    return X, Y


def mixture_dataset(num_samples, positive_mean=[0, 0], negative_mean=[0, 1], positive_cov=[0.5, 0.5],
                    negative_cov=[0.5, 0.5], **kwargs):
    X, y = get_data_from_mixture_distribution(
        num_samples=num_samples,
        positive_component_rvs=partial(multivariate_normal.rvs, positive_mean, positive_cov),
        negative_component_rvs=partial(multivariate_normal.rvs, negative_mean, negative_cov),
    )
    return X, y


def blobs_dataset(num_samples, **kwargs):
    return make_blobs(
        n_samples=num_samples,
        centers=[[0, 0], [1, 0.1]],
        n_features=2,
        random_state=0,
        cluster_std=0.5,
    )


def breast_cancer_dataset(num_samples, **kwargs):
    breast_cancer = load_breast_cancer()
    idx = np.random.randint(0, 500, num_samples)
    X = breast_cancer.data[:, [7, 9]][idx]
    y = breast_cancer.target[idx]
    return X, y


def xor_dataset(num_samples, var=1, **kwargs):
    noise = np.random.normal(0, var, (num_samples, 2))
    X = np.random.randint(0, 2, (num_samples, 2))
    y = np.abs(X[:, 0] - X[:, 1])
    X = X + noise
    return X, y


def xor_proba(X, var):
    def proba_point(p_x, p_y, X, var):
        d_x = (X[:, 0] - p_x)
        d_y = (X[:, 1] - p_y)
        d_2 = (d_x ** 2 + d_y ** 2) / (var ** 2)
        return np.exp(-d_2 / 2) / (var * 2 * np.pi)

    proba = np.zeros((len(X), 2))
    for point in [(0, 0, 0), (0, 1, 1), (1, 0, 1), (1, 1, 0)]:
        proba[:, point[2]] += proba_point(point[0], point[1], X, var)
    return proba[:, 1] / (proba.sum(axis=1))


DATASETS = {
    "Breast Cancer": breast_cancer_dataset,
    "Corner": generate_corner_dataset,
    "Blobs": blobs_dataset,
    "gaussian_mixture": mixture_dataset,
    "xor": xor_dataset
}


def load_dataset(
        dataset_name: str | None, num_samples: int = 100, **kwargs
) -> (pd.DataFrame, pd.Series):
    if dataset_name in DATASETS:
        return DATASETS[dataset_name](num_samples, **kwargs)
    else:
        raise ValueError(f"Unknown dataset: {dataset_name}")



import uuid
from os import path
import json
import urllib
import logging

import antakia
from importlib.resources import files

logger = logging.getLogger(__name__)


class MetaData:
    metadata_file = str(files("antakia").joinpath("assets/metadata.json"))

    def __init__(self):
        self.latest_version = self.get_latest_version()
        self.current_version = antakia.__version__
        self.load_file()

    def load_file(self):
        self.run_id = str(uuid.uuid4())
        try:
            metadata: dict = json.loads(open(self.metadata_file, "r").read()) if path.exists(self.metadata_file) else 0
            if isinstance(metadata, int):
                self.counter = metadata
                self.last_checked_version = None
                self.user_id = uuid.getnode()
                self.install_id = str(uuid.uuid4())
            else:
                self.counter = metadata.get('counter')
                self.last_checked_version = metadata.get('last_checked_version')
                self.user_id = metadata.get('user_id', uuid.getnode())
                self.install_id = metadata.get('install_id', str(uuid.uuid4()))
        except Exception:
            self.counter = 0
            self.last_checked_version = None
            self.user_id = uuid.getnode()
            self.install_id = str(uuid.uuid4())

    def update(self):
        self.counter += 1
        self.latest_version = self.get_latest_version()
        self.current_version = antakia.__version__

    def get_latest_version(self):
        # Check pypi for the latest version number
        try:
            contents = urllib.request.urlopen('https://pypi.org/pypi/antakia/json').read()
            data = json.loads(contents)
            latest_version = data['info']['version']
            return latest_version
        except:
            return None

    def is_latest_version(self):
        if self.latest_version:
            return self.latest_version <= self.current_version
        return True

    def save(self):
        try:
            with open(self.metadata_file, "w") as f:
                f.write(json.dumps({
                    'counter': self.counter,
                    'last_checked_version': self.latest_version,
                    'user_id': self.user_id,
                    'install_id': self.install_id,
                }))
        except:
            pass

    def start(self):
        self.load_file()
        self.update()
        self.save()


metadata = MetaData()



import os

ATK_DEFAULT_EXPLANATION_METHOD = int(os.environ.get('DEFAULT_EXPLANATION_METHOD', 1))
ATK_DEFAULT_DIMENSION = int(os.environ.get('DEFAULT_VS_DIMENSION', 2))
ATK_DEFAULT_PROJECTION = 'PaCMAP'

ATK_INIT_FIG_WIDTH = int(os.environ.get('INIT_FIG_WIDTH', 1800))
ATK_MAX_DOTS = int(os.environ.get('MAX_DOTS', 5000))

# Rule format
ATK_USE_INTERVALS_FOR_RULES = os.environ.get('USE_INTERVALS_FOR_RULES', 'True') == 'True'
ATK_MAX_RULES_DESCR_LENGTH = int(os.environ.get('MAX_RULES_DESCR_LENGTH', 200))

ATK_SHOW_LOG_MODULE_WIDGET = os.environ.get('SHOW_LOG_MODULE_WIDGET', 'False') == 'True'

# Auto cluster
ATK_MIN_POINTS_NUMBER = 100

ATK_SEND_LOG =  not (os.environ.get('ATK_SEND_LOG', 'True') == '0')



import ipyvuetify as v
import requests
from ipywidgets import widgets
import webbrowser
from importlib.resources import files
from antakia.gui.helpers.metadata import metadata


class TopBar:
    def __init__(self):
        # We count the number of times this GUI has been initialized

        self._build_widget()

    def _build_widget(self):
        star_btn = v.Btn(color='primary', children=['Star Antakia'])
        star_btn.on_event('click', self.open_web)

        self.star_dialog = v.Dialog(
            width='500',
            v_model=False,
            children=[
                v.Card(children=[
                    v.CardTitle(class_='headline gray lighten-2', primary_title=True, children=[
                        "Do you like AntakIA 😀 ?"
                    ]),
                    v.CardText(children=[
                        v.Html(
                            tag="p",
                            children=["Please star us on Github if you like our work!"],
                        ),
                        star_btn
                    ])
                ])
            ]
        )

        self.star_dialog.on_event('keydown.stop', lambda *args: None)  # close dialog on escape

        self.logo = v.Sheet(
            children=[
                widgets.Image(
                    value=self.get_logo(),
                    height=str(864 / 20) + "px",
                    width=str(3839 / 20) + "px",
                ),
                self.star_dialog
            ],
        )

        self.widget = v.AppBar(  # Top bar # 0
            class_="white",
            children=[
                self.logo,
                v.Sheet(children=self.get_version_text()),
                v.Sheet(class_='flex-fill align-stretch'),  # 02
                v.Sheet(children=[v.Menu(  # 03 # Menu for the figure width
                    v_slots=[
                        {
                            "name": "activator",
                            "variable": "props",
                            "children":
                                v.Btn(
                                    v_on="props.on",
                                    icon=True,
                                    size="x-large",
                                    children=[v.Icon(children=["mdi-tune"])],
                                    class_="ma-2 pa-3",
                                    elevation="0",
                                ),
                        }
                    ],
                    children=[
                        v.Card(  # 030 parameters menu
                            class_="pa-4",
                            rounded=True,
                            children=[],
                            min_width="500",
                        )
                    ],
                    v_model=False,
                    close_on_content_click=False,
                    offset_y=True,
                )]),  # End V.Menu
            ],  # End AppBar children
        )  # End AppBar

        self.logo.on_event('click', self.open_web)

    def open_web(self, *args):
        webbrowser.open('https://github.com/AI-vidence/antakia')

    def open(self):
        self.star_dialog.v_model = True

    def get_logo(self):
        try:
            url = 'https://drive.ai-vidence.com/s/6PQaTGEGD6iXHEp/download/antakia_horizontal.png'
            response = requests.get(url)
            if response.status_code < 300:
                return response.content
        except:
            pass
        file = files("antakia").joinpath("assets/logo_antakia_horizontal.png")  # type: ignore
        return open(file, 'rb').read()

    def get_version_text(self):
        current_version = metadata.current_version
        if metadata.is_latest_version():
            return 'v' + current_version
        else:
            return 'v' + current_version + ' - a new version is available !'



import time

import pandas as pd
import ipyvuetify as v

from antakia import config
from antakia_core.explanation.explanations import compute_explanations, ExplanationMethod
from antakia.gui.helpers.progress_bar import ProgressBar
from antakia.utils.stats import stats_logger, log_errors


class ExplanationValues:
    """
    Widget to manage explanation values
    in charge on computing them when necessary
    """
    available_exp = ['Imported', 'SHAP', 'LIME']

    def __init__(self, X: pd.DataFrame, y: pd.Series, model, task_type, on_change_callback: callable,
                 disable_gui: callable, X_exp=None):
        """

        Parameters
        ----------
        X: original train DataFrame
        y: target variable
        model: customer model
        on_change_callback: callback to notify explanation change
        X_exp: user provided explanations
        """
        self.widget = None
        self.X = X
        self.y = y
        self.model = model
        self.task_type = task_type
        self.on_change_callback = on_change_callback
        self.disable_gui = disable_gui
        self.initialized = False

        # init dict of explanations
        self.explanations: dict[str, pd.DataFrame | None] = {
            exp: None for exp in self.available_exp
        }

        if X_exp is not None:
            self.explanations[self.available_exp[0]] = X_exp

        # init selected explanation
        if X_exp is not None:
            self.current_exp = self.available_exp[0]
        else:
            self.current_exp = self.available_exp[1]
        stats_logger.log('exp_method_init', {'exp_method': self.current_exp})
        self._build_widget()

    def _build_widget(self):
        self.widget = v.Row(children=[
            v.Select(  # Select of explanation method
                label="Explanation method",
                items=[
                    {"text": "Imported", "disabled": True},
                    {"text": "SHAP", "disabled": True},
                    {"text": "LIME", "disabled": True},
                ],
                class_="ml-2 mr-2",
                style_="width: 15%",
                disabled=False,
            ),
            v.ProgressCircular(  # exp menu progress bar
                class_="ml-2 mr-2 mt-2",
                indeterminate=False,
                color="grey",
                width="6",
                size="35",
            )
        ])
        # refresh select menu
        self.update_explanation_select()
        self.get_explanation_select().on_event("change", self.explanation_select_changed)
        # set up callback
        self.get_progress_bar().reset_progress_bar()

    def initialize(self, progress_callback):
        """
        initialize class (compute explanation if necessary)
        Parameters
        ----------
        progress_callback : callback to notify progress

        Returns
        -------

        """
        if not self.has_user_exp:
            # compute explanation if not provided
            self.compute_explanation(config.ATK_DEFAULT_EXPLANATION_METHOD, progress_callback)
        # ensure progress is at 100%
        progress_callback(100, 0)
        self.initialized = True

    @property
    def current_exp_df(self) -> pd.DataFrame:
        """
        currently selected explanation projected values instance
        Returns
        -------

        """
        return self.explanations[self.current_exp]

    @property
    def has_user_exp(self) -> bool:
        """
        has the user provided an explanation
        Returns
        -------

        """
        return self.explanations[self.available_exp[0]] is not None

    def update_explanation_select(self):
        """
        refresh explanation select menu
        Returns
        -------

        """
        exp_values = []
        for exp in self.available_exp:
            if exp == 'Imported':
                exp_values.append({
                    "text": exp,
                    'disabled': self.explanations[exp] is None
                })
            else:
                exp_values.append({
                    "text": exp + (' (compute)' if self.explanations[exp] is None else ''),
                    'disabled': False
                })
        self.get_explanation_select().items = exp_values
        self.get_explanation_select().v_model = self.current_exp

    def get_progress_bar(self):
        progress_widget = self.widget.children[1]
        progress_bar = ProgressBar(progress_widget)
        return progress_bar

    def get_explanation_select(self):
        """
        returns the explanation select menu
        Returns
        -------

        """
        return self.widget.children[0]

    def compute_explanation(self, explanation_method: int, progress_bar: callable):
        """
        compute explanation and refresh widgets (select the new explanation method)
        Parameters
        ----------
        explanation_method: desired explanation
        progress_bar : progress bar to notify progress to

        Returns
        -------

        """
        t = time.time()
        self.disable_gui(True)
        # We compute proj for this new PV :
        x_exp = compute_explanations(self.X, self.model, explanation_method, self.task_type, progress_bar)
        pd.testing.assert_index_equal(x_exp.columns, self.X.columns)

        # update explanation
        self.explanations[self.available_exp[explanation_method]] = x_exp
        # refresh front
        self.update_explanation_select()
        self.disable_gui(False)
        stats_logger.log('compute_explanation',
                         {'exp_method': explanation_method,
                          'compute_time': time.time() - t})

    def disable_selection(self, is_disabled: bool):
        """
        disable widgets
        Parameters
        ----------
        is_disabled = should disable ?

        Returns
        -------

        """
        self.get_explanation_select().disabled = is_disabled

    @log_errors
    def explanation_select_changed(self, widget, event, data):
        """
        triggered on selection of new explanation by user
        explanation has already been computed (the option is enabled in select)
        Parameters
        ----------
        widget
        event
        data: explanation name

        Returns
        -------

        Called when the user chooses another dataframe
        """
        stats_logger.log('exp_method_changed', {'selected': data})
        if not isinstance(data, str):
            raise KeyError('invalid explanation')
        data = data.replace(' ', '').replace('(compute)', '')
        self.current_exp = data

        if self.explanations[self.current_exp] is None:
            exp_method = ExplanationMethod.explain_method_as_int(self.current_exp)
            progress_bar = self.get_progress_bar()
            self.compute_explanation(exp_method, progress_bar)

        self.on_change_callback(self.current_exp_df)



from logging import Handler, Logger, DEBUG, Formatter, INFO

import ipywidgets as widgets
from IPython.core.display_functions import display
from ipywidgets import Widget

from antakia import config as config


class OutputWidgetHandler(Handler):
    """Custom logging handler sending logs to an output widget"""

    def __init__(self, height: int, *args, **kwargs):
        super(OutputWidgetHandler, self).__init__(*args, **kwargs)
        layout = {"width": "100%", "height": str(height) + "px", "border": "1px solid black", "overflow_y": "auto"}
        self.out = widgets.Output(layout=layout)

    def emit(self, record):
        """Overload of logging.Handler method"""
        formatted_record = self.format(record)
        new_output = {
            "name": "stdout",
            "output_type": "stream",
            "text": formatted_record + "\n",
        }
        self.out.outputs = (new_output,) + self.out.outputs

    def show_logs(self):
        """Show the logs"""
        display(self.out)

    def clear_logs(self):
        """Clear the current logs"""
        self.out.clear_output()


def conf_logger(logger: Logger, height: int = 160) -> Handler:
    if config.ATK_SHOW_LOG_MODULE_WIDGET:
        logger.setLevel(DEBUG)
        handler = OutputWidgetHandler(height)
        handler.setFormatter(Formatter('%(asctime)s-%(levelname)s:%(module)s|%(lineno)s:: %(message)s'))
        logger.addHandler(handler)
        handler.clear_logs()
        handler.show_logs()
    else:
        logger.setLevel(INFO)


def wrap_repr(widget: Widget, size: int = 200) -> str:
    text = widget.__repr__()
    if widget.layout is None:
        text += " Layout is None !"
    else:
        text += " Visibility : " + widget.layout.visibility
    s_wrap_list = textwrap.wrap(text, size)
    return '\n'.join(s_wrap_list)



from functools import partial

import ipyvuetify as v

from antakia import config
from antakia.utils.stats import log_errors, stats_logger


class DimSwitch:
    def __init__(self, update_callback):
        self.update_callback = partial(update_callback,self)
        self._build_widget()

    def _build_widget(self):
        self.widget = v.Switch(  # 100 # Dimension switch
            v_on='tooltip.on',
            class_="ml-6 mr-2",
            v_model=config.ATK_DEFAULT_DIMENSION == 3,
            label="2D/3D",
        )

        self.widget.on_event("change", self.switch_dimension)

    @log_errors
    def switch_dimension(self, widget=None, event=None, data=None):
        dim = 3 if data else 2
        stats_logger.log('dim_changed', {'dim':dim})
        self.update_callback(dim)



import ipyvuetify as v
from ipywidgets import Layout, widgets
from importlib.resources import files

from antakia.gui.helpers.progress_bar import ProgressBar, MultiStepProgressBar


class SplashScreen:
    def __init__(self, X):
        self.X = X
        self._build_widget()

    def _build_widget(self):
        exp_progressbar_wgt = v.ProgressLinear(  # 110
            style_="width: 80%",
            class_="py-0 mx-5",
            v_model=0,
            color="primary",
            height="15",
            striped=True,
        )
        self.exp_progressbar = ProgressBar(
            exp_progressbar_wgt,
            unactive_color="light blue",
            reset_at_end=False
        )

        proj_progressbar_wgt = v.ProgressLinear(  # 110
            style_="width: 80%",
            class_="py-0 mx-5",
            v_model=0,
            color="primary",
            height="15",
            striped=True,
        )
        self.proj_progressbar = MultiStepProgressBar(
            proj_progressbar_wgt,
            steps=2,
            unactive_color="light blue",
            reset_at_end=False
        )

        self.exp_txt = v.TextField(  # 120
            variant="plain",
            v_model="",
            readonly=True,
            class_="mt-0 pt-0",
        )
        self.proj_txt = v.TextField(  # 120
            variant="plain",
            v_model="",
            readonly=True,
            class_="mt-0 pt-0",
        )

        self.widget = v.Layout(
            class_="flex-column align-center justify-center",
            children=[
                widgets.Image(  # 0
                    value=widgets.Image._load_file_value(
                        files("antakia").joinpath("assets/logo_antakia.png")
                    ),
                    layout=Layout(width="230px"),
                ),
                v.Row(  # 1
                    style_="width:85%;",
                    children=[
                        v.Col(  # 10
                            children=[
                                v.Html(
                                    tag="h3",
                                    class_="mt-2 text-right",
                                    children=["Computation of explanation values"],
                                )
                            ]
                        ),
                        v.Col(  # 11
                            class_="mt-3",
                            children=[
                                exp_progressbar_wgt
                            ],
                        ),
                        v.Col(  # #12
                            children=[
                                self.exp_txt
                            ]
                        ),
                    ],
                ),
                v.Row(  # 2
                    style_="width:85%;",
                    children=[
                        v.Col(  # 20
                            children=[
                                v.Html(
                                    tag="h3",
                                    class_="mt-2 text-right",
                                    children=["Computation of dimension reduction values"],
                                )
                            ]
                        ),
                        v.Col(  # 21
                            class_="mt-3",
                            children=[
                                proj_progressbar_wgt
                            ],
                        ),
                        v.Col(  # 22
                            children=[
                                self.proj_txt
                            ]
                        ),
                    ],
                ),
            ],
        )

    def set_exp_msg(self, msg):
        self.exp_txt.v_model = msg

    def set_proj_msg(self, msg):
        self.proj_txt.v_model = msg



from __future__ import annotations

import pandas as pd
import numpy as np
from plotly.graph_objects import FigureWidget, Scattergl, Scatter3d
import ipyvuetify as v
from sklearn.neighbors import KNeighborsClassifier

from antakia_core.data_handler.region import Region, RegionSet

import antakia_core.utils.utils as utils
import antakia.config as config

import logging as logging
from antakia.utils.logging_utils import conf_logger
from antakia.utils.stats import log_errors, stats_logger

logger = logging.getLogger(__name__)
conf_logger(logger)


class FigureDisplay:
    """
    A FigureDisplay objet manages all operation on a scatter plot 
    This class is only responsible for displaying the provided data

    It can display in 3 or 2 dimensions.

    Attributes :

    """

    # Trace indexes : 0 for values, 1 for rules, 2 for regions
    NUM_TRACES = 4
    VALUES_TRACE = 0
    RULES_TRACE = 1
    REGIONSET_TRACE = 2
    REGION_TRACE = 3

    @staticmethod
    def trace_name(trace_id: int) -> str:
        """
        get trace name from id
        Parameters
        ----------
        trace_id : int

        Returns
        -------

        """
        if trace_id == FigureDisplay.VALUES_TRACE:
            return 'values trace'
        elif trace_id == FigureDisplay.RULES_TRACE:
            return 'rules trace'
        elif trace_id == FigureDisplay.REGIONSET_TRACE:
            return 'regionset trace'
        elif trace_id == FigureDisplay.REGION_TRACE:
            return 'region trace'
        else:
            return "unknown trace"

    def __init__(
        self,
        X: pd.DataFrame | None,
        y: pd.Series,
        selection_changed: callable,
        space: str
    ):
        """

        Parameters
        ----------
        X: data to display, should be 2 or 3D 
        y: target value (default color)
        selection_changed : callable called when a selection changed
        """
        # current active trace
        self.active_trace = 0
        # mask of value to display to limit points on graph
        self._mask = None
        # callback to notify gui that the selection has changed
        self.selection_changed = selection_changed
        self.X = X
        self.y = y

        self.space = space

        # Now we can init figure
        self.widget = v.Container()
        self.widget.class_ = "flex-fill"

        # is graph selectable
        self._selection_mode = 'lasso'
        # current selection
        if X is not None:
            self._current_selection = utils.boolean_mask(self.X, True)
        else:
            self._current_selection = None
        # is this selection first since last deselection ?
        self.first_selection = True

        # traces to show
        self._visible = [True, False, False, False]
        # trace_colors
        self._colors: list[pd.Series | None] = [None, None, None, None]

        # figures
        self.figure_2D = self.figure_3D = None
        # is the class fully initialized
        self.initialized = False

    @property
    def figure(self):
        if self.dim == 2:
            return self.figure_2D
        else:
            return self.figure_3D

    @figure.setter
    def figure(self, value):
        if self.dim == 2:
            self.figure_2D = value
        else:
            self.figure_3D = value

    @property
    def dim(self):
        if self.X is None:
            return config.ATK_DEFAULT_DIMENSION
        return self.X.shape[1]

    @property
    def current_selection(self):
        if self._current_selection is None:
            self._current_selection = utils.boolean_mask(self.X, True)
        return self._current_selection

    @current_selection.setter
    def current_selection(self, value):
        self._current_selection = value

    def initialize(self, X: pd.DataFrame = None):
        """
        inital computation called at startup, after init to compute required values
        Parameters
        ----------
        X : data to display, if NOne use the one provided during init

        Returns
        -------

        """
        if X is not None:
            self.X = X
            # compute X if needed
            self.get_X(masked=True)
        self.initialized = True

    # ---- display Methods ------

    def disable_selection(self, is_disabled: bool):
        """
        enable/disable selection on graph
        Parameters
        ----------
        is_disabled

        Returns
        -------

        """

        self._selection_mode = False if is_disabled else "lasso"
        if self.dim == 2 and self.figure is not None:
            self.figure.update_layout(
                dragmode=self._selection_mode
            )

    def _show_trace(self, trace_id: int, show: bool):
        """
        show/hide trace
        Parameters
        ----------
        trace_id : trace to change
        show : show/hide

        Returns
        -------

        """
        self._visible[trace_id] = show
        self.figure.data[trace_id].visible = show

    def display_rules(self, selection_mask: pd.Series, rules_mask: pd.Series = None):
        """
        display a rule vs a selection
        Parameters
        ----------
        selection_mask: boolean series of selected points
        rules_mask: boolean series of rule validating points

        Returns
        -------

        """
        if selection_mask.all():
            selection_mask = ~selection_mask
        if rules_mask is None:
            rules_mask = selection_mask
        color, _ = utils.get_mask_comparison_color(rules_mask, selection_mask)

        self._colors[self.RULES_TRACE] = color
        self._display_zones(self.RULES_TRACE)

    def display_regionset(self, region_set: RegionSet):
        """
        display a region set, each region in its color
        Parameters
        ----------
        region_set

        Returns
        -------

        """
        self._colors[self.REGIONSET_TRACE] = region_set.get_color_serie()
        self._display_zones(self.REGIONSET_TRACE)

    def display_region(self, region: Region):
        """
        display a single region
        Parameters
        ----------
        region

        Returns
        -------

        """
        rs = RegionSet(self.X)
        rs.add(region)
        self._colors[self.REGION_TRACE] = rs.get_color_serie()
        self._display_zones(self.REGION_TRACE)

    def _display_zones(self, trace=None):
        """
        refresh provided trace or all trace if None
        do not alter visibility
        Parameters
        ----------
        trace

        Returns
        -------

        """

        if trace is None:
            for trace_id in range(self.NUM_TRACES):
                self.refresh_trace(trace_id=trace_id)
        else:
            self.refresh_trace(trace_id=trace)

    def set_color(self, color: pd.Series, trace_id: int):
        """
        set the provided color as the scatter point color on the provided trace id
        do not alter show/hide trace
        Parameters
        ----------
        color
        trace_id

        Returns
        -------

        """
        self._colors[trace_id] = color
        self.refresh_trace(trace_id)

    def refresh_trace(self, trace_id: int):
        """
        refresh the provided trace id
        do not alter show/hide trace

        Parameters
        ----------
        trace_id

        Returns
        -------

        """
        if self.figure is None:
            self.create_figure()
        else:
            colors = self._colors[trace_id]
            if colors is None:
                colors = self.y
            colors = colors[self.mask]
            with self.figure.batch_update():
                self.figure.data[trace_id].marker.color = colors

    def update_X(self, X: pd.DataFrame):
        """
        changes the underlying data - update the data used in display and dimension is necessary
        Parameters
        ----------
        X : data to display

        Returns
        -------

        """
        self.X = X
        self.redraw()

    def selection_to_mask(self, row_numbers: list[int]):
        """

        extrapolate selected row_numbers to full dataframe and return the selection mask on all the dataframe

        Parameters
        ----------
        row_numbers

        Returns
        -------

        """
        selection = utils.rows_to_mask(self.X[self.mask], row_numbers)
        if not selection.any() or selection.all():
            return utils.boolean_mask(self.get_X(masked=False), selection.mean())
        if self.mask.all():
            return selection
        X_train = self.get_X(masked=True)
        knn = KNeighborsClassifier().fit(X_train, selection)
        X_predict = self.get_X(masked=False)
        guessed_selection = pd.Series(knn.predict(X_predict), index=X_predict.index)
        # KNN extrapolation
        return guessed_selection.astype(bool)

    @log_errors
    def _selection_event(self, trace, points, *args):
        """
        callback triggered by selection on graph
        selects points and update display on both hde (calls selection changed)
        deselects if no points selected

        update selection taking intersection
        Parameters
        ----------
        trace
        points
        args

        Returns
        -------

        """
        self.first_selection |= self.current_selection.all()
        stats_logger.log('hde_selection', {'first_selection': self.first_selection, 'space': self.space})
        self.current_selection &= self.selection_to_mask(points.point_inds)
        self.display_rules(self.current_selection)
        if self.current_selection.any():
            self.create_figure()
            self.selection_changed(self, self.current_selection)
        else:
            self._deselection_event(rebuild=True)

    @log_errors
    def _deselection_event(self, *args, rebuild=False):
        """
        clear selection -- called by deselection on graph
        synchronize hdes

        Parameters
        ----------
        args
        rebuild

        Returns
        -------

        """
        stats_logger.log('hde_deselection', {'first_selection': self.first_selection, 'space': self.space})
        # We tell the GUI
        self.first_selection = False
        self.current_selection = utils.boolean_mask(self.X, True)
        self.selection_changed(self, self.current_selection)
        self.display_rules(~self.current_selection, ~self.current_selection)
        if rebuild:
            self.create_figure()
        else:
            self.display_selection()

    def set_selection(self, new_selection_mask: pd.Series):
        """
        update selection from mask
        no update_callback
        Called by tne UI when a new selection occurred on the other HDE
        Parameters
        ----------
        new_selection_mask

        Returns
        -------

        """

        if self.current_selection.all() and new_selection_mask.all():
            # New selection is empty. We already have an empty selection : nothing to do
            return

        # selection event
        self.current_selection = new_selection_mask
        self.display_rules(self.current_selection, self.current_selection)
        self.display_selection()
        return

    def display_selection(self):
        """
        display selection on figure
        Returns
        -------

        """
        if self.dim == 2:
            fig = self.figure.data[0]
            fig.update(selectedpoints=utils.mask_to_rows(self.current_selection[self.mask]))
            fig.selectedpoints = utils.mask_to_rows(self.current_selection[self.mask])

    @property
    def mask(self) -> pd.Series:
        """
        mask should be applied on each display (x,y,z,color, selection)
        """
        if self._mask is None:
            self._mask = pd.Series([False] * len(self.X), index=self.X.index)
            limit = config.ATK_MAX_DOTS
            if len(self.X) > limit:
                indices = np.random.choice(self.X.index, size=limit, replace=False)
                self._mask.loc[indices] = True
            else:
                self._mask.loc[:] = True
        return self._mask

    def create_figure(self):
        """
        Builds the FigureWidget for the given dimension
        """
        x = y = z = None

        if self.X is not None:
            proj_values = self.get_X(masked=True)

        hde_marker = {'color': self.y, 'colorscale': "Viridis"}
        if self.dim == 3:
            hde_marker['size'] = 2

        fig_args = {
            'x': proj_values[0],
            'y': proj_values[1],
            'mode': "markers",
            'marker': hde_marker,
            'customdata': self.y[self.mask],
            'hovertemplate': "%{customdata:.3f}",
        }
        if self.dim == 3:
            fig_args['z'] = proj_values[2]
            fig_builder = Scatter3d
        else:
            fig_builder = Scattergl

        self.figure = FigureWidget(data=[fig_builder(**fig_args)])  # Trace 0 for dots
        self.figure.add_trace(fig_builder(**fig_args))  # Trace 1 for rules
        self.figure.add_trace(fig_builder(**fig_args))  # Trace 2 for region set
        self.figure.add_trace(fig_builder(**fig_args))  # Trace 3 for region

        self.figure.update_layout(dragmode=self._selection_mode)
        self.figure.update_traces(
            selected={"marker": {"opacity": 1.0}},
            unselected={"marker": {"opacity": 0.1}},
            selector={'type': "scatter"}
        )
        self.figure.update_layout(
            autosize=True,
            margin={
                't': 0,
                'b': 0,
                'l': 0,
                'r': 0
            },
        )
        self.figure._config = self.figure._config | {"displaylogo": False}
        self.figure._config = self.figure._config | {'displayModeBar': True}
        # We don't want the name of the trace to appear :
        for trace_id in range(len(self.figure.data)):
            self.figure.data[trace_id].showlegend = False
            self._show_trace(trace_id, self._visible[trace_id])
            self.refresh_trace(trace_id)
        self.display_selection()

        if self.dim == 2:
            # selection only on trace 0
            self.figure.data[0].on_selection(self._selection_event)
            self.figure.data[0].on_deselect(self._deselection_event)
        self.widget.children = [self.figure]


    def redraw(self):
        """
        redraw all traces, without recreating figure
        Returns
        -------

        """
        if self.figure is None:
            self.create_figure()
        projection = self.get_X(masked=True)

        with self.figure.batch_update():
            for trace_id in range(len(self.figure.data)):
                self.figure.data[trace_id].x = projection[0]
                self.figure.data[trace_id].y = projection[1]
                if self.dim == 3:
                    self.figure.data[trace_id].z = projection[2]
                self.figure.data[trace_id].showlegend = False
                self._show_trace(trace_id, self._visible[trace_id])
                self.refresh_trace(trace_id)
        self.widget.children = [self.figure]

    def get_X(self, masked: bool) -> pd.DataFrame | None:
        """

        return current projection value
        its computes it if necessary - progress is published in the callback

        Parameters
        ----------
        masked

        Returns
        -------

        """
        if masked and self.X is not None:
            return self.X.loc[self.mask]
        return self.X

    def set_tab(self, tab):
        """
        show/hide trace depending on tab
        Parameters
        ----------
        tab

        Returns
        -------

        """
        self.disable_selection(tab > 1)
        self._show_trace(self.VALUES_TRACE, tab == 0)
        self._show_trace(self.RULES_TRACE, tab == 1)
        self._show_trace(self.REGIONSET_TRACE, tab == 2)
        self._show_trace(self.REGION_TRACE, tab == 3)
        # and it's the only place where selection is allowed
        self.active_trace = tab



import pandas as pd
from antakia_core.data_handler.projected_values import ProjectedValues


class ProjectedValueBank:
    def __init__(self, y: pd.Series):
        self.projected_values = {}
        self.y = y

    def get_projected_values(self, X: pd.DataFrame) -> ProjectedValues:
        if id(X) not in self.projected_values:
            self.projected_values[id(X)] = ProjectedValues(X, self.y)
        return self.projected_values[id(X)]



import math

import ipyvuetify as v
import pandas as pd


class ProgressBar:
    def __init__(self, widget, indeterminate: bool = False, active_color='blue', unactive_color='grey',
                 reset_at_end=True):
        """
        generic progress bar update
        Parameters
        ----------
        widget : widget element
        indeterminate : whether the progress is indeterminate or finegrained
        active_color
        unactive_color
        reset_at_end : should we reset widget when 100% is reached
        """
        self.active_color = active_color
        self.unactive_color = unactive_color
        assert isinstance(widget, (v.ProgressLinear, v.ProgressCircular))
        self.widget = widget
        self.indeterminate = indeterminate
        self.progress = 0
        self.reset_at_end = reset_at_end

    def update(self, progress: float, time_elapsed=None):
        """

        Parameters
        ----------
        progress: progress value between 0 and 100
        time_elapsed : duration since start

        Returns
        -------

        """
        self.progress = progress
        self.widget.color = self.active_color
        self.widget.indeterminate = self.indeterminate

        if math.ceil(progress) >= 100 and self.reset_at_end:
            self.reset_progress_bar()

    def reset_progress_bar(self):
        self.progress = 100
        self.widget.indeterminate = False
        self.widget.color = self.unactive_color

    def __call__(self, *args, **kwargs):
        return self.update(*args, **kwargs)

    @property
    def progress(self):
        return self.widget.v_model

    @progress.setter
    def progress(self, value: float):
        if value is None or pd.isna(value):
            self.widget.indeterminate = True
        else:
            if self.indeterminate and value <= 99:
                self.widget.indeterminate = True
            else:
                self.widget.indeterminate = False
            self.widget.v_model = value


class MultiStepProgressBar:
    def __init__(self, widget, steps=1, active_color='blue', unactive_color='grey', reset_at_end=True):
        """
        generic progress bar update

        Parameters
        ----------
        widget : widget element
        steps
        active_color
        unactive_color
        reset_at_end : should we reset widget when 100% is reached
        """
        self.steps = steps
        self.widget = widget
        self.active_color = active_color
        self.unactive_color = unactive_color
        self.widget.v_model = 0
        self.reset_at_end = reset_at_end

    def get_update(self, step):
        """
        returns the progress updater for the provided step
        Parameters
        ----------
        step

        Returns
        -------

        """
        if step == 0 or step > self.steps:
            raise ValueError('step should be between 1 and self.steps')

        def update_ac_progress_bar(progress: float, duration: float):
            """
            Called by the AutoCluster to update the progress bar
            """
            self.widget.color = self.active_color
            progress = round(((step - 1) * 100 + progress) / self.steps)
            self.set_progress(progress)

        return update_ac_progress_bar

    def set_progress(self, progress: float):
        """
        force progress value
        Parameters
        ----------
        progress

        Returns
        -------

        """
        self.widget.v_model = progress
        if progress >= 100 and self.reset_at_end:
            self.reset_progress_bar()

    @property
    def progress(self):
        return self.widget.v_model

    @progress.setter
    def progress(self, value):
        self.set_progress(value)

    def reset_progress_bar(self):
        self.progress = 0
        self.widget.color = self.unactive_color



import time

import pandas as pd

from antakia_core.compute.dim_reduction.dim_reduc_method import DimReducMethod
from antakia_core.compute.dim_reduction.dim_reduction import dim_reduc_factory
from antakia_core.data_handler.projected_values import Proj, ProjectedValues

from antakia import config
from antakia.gui.high_dim_exp.projected_value_bank import ProjectedValueBank
from antakia.gui.helpers.progress_bar import ProgressBar
from ipywidgets import widgets
import ipyvuetify as v

from antakia_core.utils import utils

from antakia.utils.stats import stats_logger, log_errors


class ProjectedValuesSelector:
    def __init__(self, pv_bank: ProjectedValueBank, update_callback: callable, space):
        self.widget = None
        self.progress_bar = None
        self.projected_value: ProjectedValues | None = None
        self._proj_params_cards = {}
        self.update_callback = update_callback
        self.pv_bank = pv_bank
        self.space = space

        self.X = None
        self.current_proj = Proj(
            DimReducMethod.dimreduc_method_as_int(config.ATK_DEFAULT_PROJECTION),
            config.ATK_DEFAULT_DIMENSION
        )

        self._build_widget()

    @property
    def current_dim(self):
        return self.current_proj.dimension

    def _build_widget(self):
        self.widget = v.Row(children=[
            v.Select(  # Selection of proj method
                label=f"Projection in the {self.space} :",
                items=DimReducMethod.dimreduc_methods_as_str_list(),
                style_="width: 15%",
                class_="ml-2 mr-2",
            ),
            v.Menu(  # proj settings
                class_="ml-2 mr-2",
                v_slots=[
                    {
                        "name": "activator",
                        "variable": "props",
                        "children": v.Btn(
                            v_on="props.on",
                            icon=True,
                            size="x-large",
                            children=[
                                v.Icon(
                                    children=["mdi-cogs"],
                                    size="large",
                                )
                            ],
                            class_="ma-2 pa-3",
                            elevation="3",
                        ),
                    }
                ],
                children=[
                    v.Card(  # 1410
                        class_="pa-4",
                        rounded=True,
                        children=[
                            widgets.VBox(  # 14100
                                [
                                    v.Slider(  # 141000
                                        class_="ma-8 pa-2",
                                        v_model=10,
                                        min=5,
                                        max=30,
                                        step=1,
                                        label="Number of neighbours",
                                        thumb_label="always",
                                        thumb_size=25,
                                    ),
                                    v.Slider(  # 141001
                                        class_="ma-8 pa-2",
                                        v_model=0.5,
                                        min=0.1,
                                        max=0.9,
                                        step=0.1,
                                        label="MN ratio",
                                        thumb_label="always",
                                        thumb_size=25,
                                    ),
                                    v.Slider(  # 141002
                                        class_="ma-8 pa-2",
                                        v_model=2,
                                        min=0.1,
                                        max=5,
                                        step=0.1,
                                        label="FP ratio",
                                        thumb_label="always",
                                        thumb_size=25,
                                    )
                                ],
                            )
                        ],
                        min_width="500",
                    )
                ],
                v_model=False,
                close_on_content_click=False,
                offset_y=True,
            ),
            v.ProgressCircular(  # progress bar
                indeterminate=True,
                color="blue",
                width="6",
                size="35",
                class_="ml-2 mr-2 mt-2",
            )
        ])
        self.progress_bar = ProgressBar(self.widget.children[2], indeterminate=True)
        self.progress_bar.update(100, 0)
        self.projection_select.on_event("change", self.projection_select_changed)
        self.build_all_proj_param_w()
        self.projection_select.on_event("change", self.projection_select_changed)

    def initialize(self, progress_callback, X: pd.DataFrame):
        self.projected_value = self.pv_bank.get_projected_values(X)
        self.get_current_X_proj(progress_callback=progress_callback)
        self.refresh()

    def refresh(self):
        self.disable(True)
        self.projection_select.v_model = DimReducMethod.dimreduc_method_as_str(
            self.current_proj.reduction_method
        )
        self.update_proj_params_menu()
        self.update_callback()
        self.disable(False)

    def update_X(self, X: pd.DataFrame):
        self.projected_value = self.pv_bank.get_projected_values(X)
        self.refresh()

    def update_dim(self, dim):
        self.current_proj = Proj(self.current_proj.reduction_method, dim)
        self.refresh()

    @property
    def projection_select(self):
        """
        get dim reduc selector
        Returns
        -------

        """
        return self.widget.children[0]

    @property
    def projection_method(self) -> int:
        """
        returns the current projection method
        Returns
        -------

        """
        if self.projection_select.v_model == '!!disabled!!':
            self.projection_select.v_model = config.ATK_DEFAULT_PROJECTION
        return DimReducMethod.dimreduc_method_as_int(
            self.projection_select.v_model
        )

    @log_errors
    def projection_select_changed(self, *args):
        """
        callback called on projection select change
        projection is computed if needed
        Parameters
        ----------
        widget
        event
        data

        Returns
        -------

        """
        self.current_proj = Proj(self.projection_method, self.current_dim)
        self.update_proj_params_menu()
        self.refresh()

    @property
    def proj_param_widget(self):
        """
        get the projection parameter widget
        Returns
        -------

        """
        return self.widget.children[1]

    def build_proj_param_widget(self, dim_reduc) -> list[v.Slider]:
        """
        build widget
        Parameters
        ----------
        dim_reduc

        Returns
        -------

        """
        parameters = dim_reduc_factory[dim_reduc].parameters()
        sliders = []
        for param, info in parameters.items():
            min_, max_, step = utils.compute_step(info['min'], info['max'])
            default_value = info['default']
            if info['type'] == int:
                step = max(round(step), 1)

            slider = v.Slider(  # 15000
                class_="ma-8 pa-2",
                v_model=default_value,
                min=float(min_),
                max=float(max_),
                step=step,
                label=param,
                thumb_label="always",
                thumb_size=25,
            )
            slider.on_event("change", self.params_changed)
            sliders.append(slider)
        return sliders

    def update_proj_param_value(self):
        parameters = self.projected_value.get_parameters(self.current_proj)['current']
        param_widget = self._proj_params_cards[self.projection_method]
        for slider in param_widget:
            slider.v_model = parameters[slider.label]

    def build_all_proj_param_w(self):
        for dim_reduc in DimReducMethod.dimreduc_methods_as_list():
            self._proj_params_cards[dim_reduc] = self.build_proj_param_widget(dim_reduc)

    @log_errors
    def params_changed(self, widget, event, data):
        """
        called when user changes a parameter value
        Parameters
        ----------
        widget: caller widget
        event -
        data : new value

        Returns
        -------

        """
        self.update_params(widget.label, data)

    def update_params(self, parameter, new_value):
        self.projected_value.set_parameters(self.current_proj,
                                            {parameter: new_value})
        self.refresh()

    def update_proj_params_menu(self):
        """
        Called at startup by the GUI
        """
        # We return
        params = self._proj_params_cards[self.projection_method]
        # We neet to set a Card, depending on the projection method
        self.proj_param_widget.children[0].children = [widgets.VBox(params)]
        self.update_proj_param_value()

    def disable_select(self, is_disabled: bool):
        self.projection_select.disabled = is_disabled

    def disable_params(self, is_disabled: bool):
        self.proj_param_widget.disabled = is_disabled

    def disable(self, is_disabled):
        params = self._proj_params_cards[self.projection_method]
        self.disable_select(is_disabled)
        # do not enable proj parama menu if there are no parameters
        is_disabled |= len(params) == 0
        self.disable_params(is_disabled)

    def get_current_X_proj(self, dim=None, progress_callback=None) -> pd.DataFrame | None:
        """
        get current project X
        Parameters
        ----------
        dim: dimension to get, if None use current
        progress_callback: callback to publish progress to, if None use default

        Returns
        -------

        """
        if dim is None:
            dim = self.current_dim
        if progress_callback is None:
            progress_callback = self.progress_bar.update
        is_present = self.projected_value.is_present(Proj(self.current_proj.reduction_method, dim))
        t = time.time()
        X = self.projected_value.get_projection(
            Proj(self.current_proj.reduction_method, dim), progress_callback
        )
        if not is_present:
            stats_logger.log('compute_projection',
                             {'projection_method': self.current_proj.reduction_method, 'dimension': dim,
                              'compute_time': time.time() - t})

        return X

    def is_computed(self, projection_method=None, dim=None) -> bool:
        if projection_method is None:
            projection_method = self.projection_method
        if dim is None:
            dim = self.current_dim
        return self.projected_value.is_present(Proj(projection_method, dim))



from __future__ import annotations

from typing import List, Dict, Any

import numpy as np
import pandas as pd

from antakia_core.utils.utils import ProblemCategory

from antakia.utils.stats import stats_logger, log_errors
from antakia.utils.checks import is_valid_model
from antakia_core.utils.variable import Variable, DataVariables
from antakia.gui.gui import GUI


class AntakIA:
    """AntakIA class. 

    Antakia instances provide data and methods to explain a ML model.

    Instance attributes
    -------------------
    X : pd.DataFrame the training dataset
    y : pd.Series the target value
    model : Model
        the model to explain
    variables : a list of Variables, describing X_list[0]
    X_test : pd.DataFrame the test dataset
    y_test : pd.Series the test target value
    score : reference scoring function
    """

    @log_errors
    def __init__(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        model,
        variables: DataVariables | List[Dict[str, Any]] | pd.DataFrame | None = None,
        X_test: pd.DataFrame = None,
        y_test: pd.Series = None,
        X_exp: pd.DataFrame | None = None,
        score: callable | str = 'auto',
        problem_category: str = 'auto'
    ):
        """
        AntakiIA constructor.

        Parameters:
            X : pd.DataFrame the training dataset
            y : pd.Series the target value
            model : Model
                the model to explain
            variables : a list of Variables, describing X_list[0]
            X_test : pd.DataFrame the test dataset
            y_test : pd.Series the test target value
            score : reference scoring function
        """
        stats_logger.log('launched', {})

        if not is_valid_model(model):
            raise ValueError(model, " should implement predict and score methods")
        X, y, X_exp = self._preprocess_data(X, y, X_exp)
        if X_test is not None:
            X_test, y_test, _ = self._preprocess_data(X_test, y_test, None)
        self.X = X
        if y.ndim > 1:
            y = y.squeeze()
        self.y = y.astype(float)

        self.X_test = X_test
        if y_test is not None and y_test.ndim > 1:
            y_test = y_test.squeeze()
        self.y_test = y_test

        self.model = model

        self.X_exp = X_exp

        self.problem_category = self._preprocess_problem_category(problem_category, model, X)
        self.score = self._preprocess_score(score, self.problem_category)

        self.set_variables(X, variables)

        self.gui = GUI(
            self.X,
            self.y,
            self.model,
            self.variables,
            self.X_test,
            self.y_test,
            self.X_exp,
            self.score,
            self.problem_category
        )
        stats_logger.log('launch_info', {'data_dim': str(self.X.shape), 'category': str(self.problem_category),
                                         'provided_exp': X_exp is not None, 'test_dataset': X_test is not None})

    def set_variables(self, X, variables):
        if variables is not None:
            if isinstance(variables, list):
                self.variables: DataVariables = Variable.import_variable_list(variables)
                if len(self.variables) != len(X.columns):
                    raise ValueError("Provided variable list must be the same length of the dataframe")
            elif isinstance(variables, pd.DataFrame):
                self.variables = Variable.import_variable_df(variables)
            else:
                raise ValueError("Provided variable list must be a list or a pandas DataFrame")
        else:
            self.variables = Variable.guess_variables(X)

    def start_gui(self) -> GUI:
        return self.gui.initialize()

    def export_regions(self):
        return self.gui.region_set

    def _preprocess_data(self, X: pd.DataFrame, y, X_exp: pd.DataFrame | None):
        if isinstance(X, np.ndarray):
            X = pd.DataFrame(X)
        if isinstance(X_exp, np.ndarray):
            X_exp = pd.DataFrame(X_exp)
        if isinstance(y, np.ndarray):
            y = pd.Series(y)

        X.columns = [str(col) for col in X.columns]
        if X_exp is not None:
            X_exp.columns = X.columns

        if X_exp is not None:
            pd.testing.assert_index_equal(X.index, X_exp.index, check_names=False)
            if X.reindex(X_exp.index).iloc[:, 0].isna().sum() != X.iloc[:, 0].isna().sum():
                raise IndexError('X and X_exp must share the same index')
        pd.testing.assert_index_equal(X.index, y.index, check_names=False)
        return X, y, X_exp

    def _preprocess_problem_category(self, problem_category: str, model, X: pd.DataFrame) -> ProblemCategory:
        if problem_category not in [e.name for e in ProblemCategory]:
            raise ValueError('Invalid problem category')
        if problem_category == 'auto':
            if hasattr(model, 'predict_proba'):
                return ProblemCategory['classification_with_proba']
            pred = self.model.predict(self.X.sample(min(100, len(self.X))))
            if len(pred.shape) > 1 and pred.shape[1] > 1:
                return ProblemCategory['classification_proba']
            return ProblemCategory['regression']
        if problem_category == 'classification':
            if hasattr(model, 'prodict_proba'):
                return ProblemCategory['classification_with_proba']
            pred = model.predict(X.sample(min(100, len(X))))
            if len(pred.shape) > 1 and pred.shape[1] > 1:
                return ProblemCategory['classification_proba']
            return ProblemCategory['classification_label_only']
        return ProblemCategory[problem_category]

    def _preprocess_score(self, score, problem_category):
        if callable(score):
            return score
        if score != 'auto':
            return score
        if problem_category == ProblemCategory.regression:
            return 'mse'
        return 'accuracy'

    def predict(self, X):
        return self.gui.region_set.predict(X)



import json
import os
import time
import traceback
from functools import wraps
from importlib.resources import files

import requests

from antakia import config
from antakia.gui.helpers.metadata import metadata


class ActivityLogger:
    log_file = files("antakia").joinpath("assets/")
    url = 'https://api.antakia.ai/'
    limit = 10
    size_limit = 1000
    send_events = ['execution_error', 'launched', 'validate_sub_model', 'auto_cluster', 'compute_explanation',
                   'validate_rules']

    def __init__(self, log_file='logs.json'):
        self.log_file = str(self.log_file.joinpath(log_file))
        self._logs = []
        if os.path.exists(self.log_file):
            with open(self.log_file, 'r') as log_file:
                for log in log_file:
                    self._logs.append(json.loads(log))
        if len(self._logs) > 0:
            self._send(force_send=True)

    def log(self, event: str, info: dict | None = None):
        """
        logs the user activity for usage statistics

        Parameters
        ----------
        event: user action
        info: mate information

        Returns
        -------

        """
        if os.environ.get('ATK_SEND_LOG', '1') == '0' or not config.ATK_SEND_LOG:
            return
        if info is None:
            info = {}
        payload = {
            'event': event,
            'log': info,
            'timestamp': time.time()
        }
        try:
            json.dumps(payload)
            self._add_to_log_queue(payload)
            self._send(force_send=payload['event'] in self.send_events)
        except:
            pass

    def _add_metadata(self, payload):
        payload['user_id'] = metadata.user_id
        payload['run_id'] = metadata.run_id
        payload['install_id'] = metadata.install_id
        payload['launch_count'] = metadata.counter
        payload['version_number'] = metadata.current_version

    def _add_to_log_queue(self, payload):
        self._logs.append(payload)
        self._add_to_disk(payload)

    def _add_to_disk(self, payload):
        try:
            with open(self.log_file, 'a') as log_file:
                log_file.write(json.dumps(payload) + '\n')
        except:
            pass

    def _clear_logs(self):
        self._logs = []
        try:
            with open(self.log_file, 'w') as log_file:
                log_file.write('')
        except:
            pass

    def _send(self, force_send=False):
        if len(self._logs) > self.limit or force_send:
            try:
                payload = {'items': self._logs}
                self._add_metadata(payload)
                response = requests.post(self.url + 'log', json=payload, timeout=10)
                if response.status_code >= 300:
                    raise ConnectionError
                self._clear_logs()
            except:
                payload = {
                    'event': 'no connection',
                    'timestamp': time.time()
                }
                self._add_to_log_queue(payload)
                if len(self._logs) > self.size_limit:
                    self._clear_logs()
                    payload = {
                        'event': 'no connection log erased',
                        'timestamp': time.time()
                    }
                    self._add_to_log_queue(payload)


stats_logger = ActivityLogger()


def log_errors(method):
    """
    decorator to log error raised in functions
    Parameters
    ----------
    method

    Returns
    -------

    """

    @wraps(method)
    def log(*args, **kw):
        try:
            res = method(*args, **kw)
        except Exception as e:
            stats_logger.log('execution_error', {
                'method': method.__qualname__,
                'traceback': analyze_traceback(e.__traceback__)
            })
            raise e
        return res

    return log


def analyze_traceback(tb: traceback):
    stack_summary = traceback.extract_tb(tb)
    # keep only antakia frames
    first_antakia_frame = 0
    for i, frame in enumerate(stack_summary):
        if '/antakia/' in frame.filename:
            first_antakia_frame = i
            break
    stack_summary = stack_summary[first_antakia_frame:]
    return [frame_to_dict(f) for f in stack_summary]


def frame_to_dict(tb_frame: traceback.FrameSummary):
    return {
        'lineno': tb_frame.lineno,
        'end_lineno': tb_frame.end_lineno,
        'line': tb_frame.line,
        'method_name': tb_frame.name,
        'filename': anonymize_filename(tb_frame.filename)
    }


def anonymize_filename(filename: str):
    return filename[filename.find('antakia'):]



from __future__ import annotations

import pandas as pd
from antakia_core.data_handler.projected_values import ProjectedValues

from antakia.gui.high_dim_exp.figure_display import FigureDisplay
from antakia.gui.high_dim_exp.projected_value_bank import ProjectedValueBank
from antakia.gui.high_dim_exp.projected_values_selector import ProjectedValuesSelector

import logging as logging
from antakia.utils.logging_utils import conf_logger

logger = logging.getLogger(__name__)
conf_logger(logger)


class HighDimExplorer:
    """
    An HighDimExplorer displays one high dim Dataframes on a scatter plot.
    This class is only responsible for displaying the provided projection
    dimensionality reduction is handled by ProjectedValueSelector

    It can display in 3 or 2 dimensions.

    Implementation details :
    It handles projections computation and switch through ProjectedValueSelector
    it can update its high dim dataframe through the update_pv method

    Attributes are mostly privates (underscorred) since they are not meant to be used outside of the class.

    Attributes :

    """

    def __init__(
        self,
        pv_bank: ProjectedValueBank,
        selection_changed: callable,
        space: str
    ):
        """

        Parameters
        ----------
        pv_bank: projected values storage
        selection_changed : callable called when a selection changed
        """
        self.pv_bank = pv_bank

        # projected values handler & widget
        self.projected_value_selector = ProjectedValuesSelector(
            pv_bank,
            self.refresh,
            space
        )

        self.figure = FigureDisplay(
            None,
            pv_bank.y,
            selection_changed,
            space
        )

        self.initialized = False

    def get_current_X_proj(self, dim=None, progress_callback=None) -> pd.DataFrame | None:
        return self.projected_value_selector.get_current_X_proj(dim, progress_callback)

    def initialize(self, progress_callback, X: pd.DataFrame):
        """
        inital computation (called at startup, after init to compute required values
        Parameters
        ----------
        progress_callback : callable to notify progress

        Returns
        -------

        """
        self.projected_value_selector.initialize(progress_callback, X)
        self.figure.initialize(self.get_current_X_proj())
        self.initialized = True

    def disable(self, disable_figure: bool, disable_projection: bool):
        """
        disable dropdown select
        Parameters
        ----------
        is_disabled: disable value

        Returns
        -------

        """
        self.figure.disable_selection(disable_figure)
        self.projected_value_selector.disable(disable_projection)

    def refresh(self, progress_callback=None):
        self.disable(True, True)
        self.figure.update_X(self.get_current_X_proj(progress_callback=progress_callback))
        self.disable(False, False)

    def update_X(self, X: pd.DataFrame, progress_callback=None):
        """
        changes the undelying projected value instance - update the data used in display
        Parameters
        ----------
        pv
        progress_callback

        Returns
        -------

        """
        self.projected_value_selector.update_X(X)
        self.refresh(progress_callback)

    @property
    def current_pv(self) -> ProjectedValues:
        return self.projected_value_selector.projected_value

    @property
    def current_X(self) -> pd.DataFrame | None:
        """
        return hde current X value (not projected)

        Returns
        -------

        """
        if self.projected_value_selector is None:
            return None  # When we're an ES HDE and no explanation have been imported nor computed yet
        return self.projected_value_selector.projected_value.X

    def set_tab(self, *args, **kwargs):
        return self.figure.set_tab(*args, **kwargs)

    def set_selection(self, *args, **kwargs):
        return self.figure.set_selection(*args, **kwargs)

    def set_dim(self, dim: int):
        self.projected_value_selector.update_dim(dim)



import pandas as pd
from antakia_core.compute.skope_rule.skope_rule import skope_rules
import ipyvuetify as v
from antakia_core.data_handler.region import Region
from antakia_core.data_handler.rules import RuleSet
from antakia_core.utils.utils import format_data

from antakia.gui.tabs.ruleswidget import RulesWidget
from antakia.gui.widget_utils import change_widget
from antakia.utils.stats import log_errors, stats_logger


class Tab1:
    def __init__(self, variables, update_callback, validate_rules_callback, X, X_exp, y):
        self.selection_changed = False
        self.region = Region(X)
        self.selection_mask = self.region.mask
        self.update_callback = update_callback
        self.validate_rules_callback = validate_rules_callback

        self.X = X
        self.X_exp = X_exp
        self.y = y

        self.variables = variables
        self.vs_rules_wgt = RulesWidget(self.X, self.y, self.variables, True, self.new_rules_defined)
        self.es_rules_wgt = RulesWidget(self.X_exp, self.y, self.variables, False)

        self._build_widget()

    def _build_widget(self):
        self.find_rules_btn = v.Btn(  # 43010 Skope button
            v_on='tooltip.on',
            class_="ma-1 primary white--text",
            children=[
                v.Icon(
                    class_="mr-2",
                    children=[
                        "mdi-axis-arrow"
                    ],
                ),
                "Find rules",
            ],
        )
        self.cancel_btn = v.Btn(  # 4302
            class_="ma-1",
            children=[
                v.Icon(
                    class_="mr-2",
                    children=[
                        "mdi-close-circle-outline"
                    ],
                ),
                "Cancel",
            ],
        )
        self.undo_btn = v.Btn(  # 4302
            class_="ma-1",
            children=[
                v.Icon(
                    class_="mr-2",
                    children=[
                        "mdi-undo"
                    ],
                ),
                "Undo",
            ],
        )
        self.validate_btn = v.Btn(  # 43030 validate
            v_on='tooltip.on',
            class_="ma-1 green white--text",
            children=[
                v.Icon(
                    class_="mr-2",
                    children=[
                        "mdi-check"
                    ],
                ),
                "Validate rules",
            ],
        )
        self.title_wgt = v.Row(children=[v.Html(  # 44000
            tag="h3",
            class_="ml-2",
            children=["Creating new region :"],
        )])
        self.selection_status_str_1 = v.Html(  # 430000
            tag="strong",
            children=["0 points"]
            # 4300000 # selection_status_str_1
        )
        self.selection_status_str_2 = v.Html(  # 43001
            tag="li",
            children=["0% of the dataset"]
            # 430010 # selection_status_str_2
        )
        self.data_table = v.DataTable(  # 432010
            v_model=[],
            show_select=False,
            headers=[
                {
                    "text": column,
                    "sortable": True,
                    "value": column,
                }
                for column in self.X.columns
            ],
            items=[],
            hide_default_footer=False,
            disable_sort=False,
        )
        self.widget = [
            self.title_wgt,
            v.Row(  # buttons row # 430
                class_="d-flex flex-row align-top mt-2",
                children=[
                    v.Sheet(  # Selection info # 4300
                        class_="ml-3 mr-3 pa-2 align-top grey lighten-3",
                        style_="width: 20%",
                        elevation=1,
                        children=[
                            v.Html(  # 43000
                                tag="li",
                                children=[
                                    self.selection_status_str_1
                                ]
                            ),
                            self.selection_status_str_2
                        ],
                    ),
                    v.Tooltip(  # 4301
                        bottom=True,
                        v_slots=[
                            {
                                'name': 'activator',
                                'variable': 'tooltip',
                                'children':
                                    self.find_rules_btn
                            }
                        ],
                        children=['Find a rule to match the selection']
                    ),
                    self.cancel_btn,
                    self.undo_btn,
                    v.Tooltip(  # 4303
                        bottom=True,
                        v_slots=[
                            {
                                'name': 'activator',
                                'variable': 'tooltip',
                                'children':
                                    self.validate_btn
                            }
                        ],
                        children=['Promote current rules as a region']
                    ),
                ]
            ),  # End Buttons row
            v.Row(  # tab 1 / row #2 : 2 RulesWidgets # 431
                class_="d-flex flex-row",
                children=[
                    self.vs_rules_wgt.widget,
                    self.es_rules_wgt.widget
                ],  # end Row
            ),
            v.ExpansionPanels(  # tab 1 / row #3 : datatable with selected rows # 432
                class_="d-flex flex-row",
                children=[
                    v.ExpansionPanel(  # 4320 # is enabled or disabled when no selection
                        children=[
                            v.ExpansionPanelHeader(  # 43200
                                class_="grey lighten-3",
                                children=["Data selected"]
                            ),
                            v.ExpansionPanelContent(  # 43201
                                children=[
                                    self.data_table
                                ],

                            ),
                        ]
                    )
                ],
            ),
        ]
        # get_widget(self.widget[2], "0").disabled = True  # disable datatable
        # We wire the click events
        self.find_rules_btn.on_event("click", self.compute_skope_rules)
        self.undo_btn.on_event("click", self.undo_rules)
        self.cancel_btn.on_event("click", self.cancel_edit)
        self.validate_btn.on_event("click", self.validate_rules)
        self.refresh_buttons()

    @property
    def valid_selection(self):
        return self.selection_mask.any() and not (self.selection_mask.all())

    @property
    def edit_type(self):
        return 'creation' if self.region.num == -1 else 'edition'

    def _update_title_txt(self):
        if self.edit_type == 'creation':
            self.title_wgt.children = [v.Html(  # 44000
                tag="h3",
                class_="ml-2",
                children=["Creating new region :"],
            )]
        else:
            region_prefix_wgt = v.Html(class_="mr-2", tag="h3", children=["Editing Region"])  # 450000
            region_chip_wgt = v.Chip(color=self.region.color, children=[str(self.region.num)], )
            self.title_wgt.children = [v.Sheet(  # 45000
                class_="ma-1 d-flex flex-row align-center",
                children=[
                    region_prefix_wgt,
                    region_chip_wgt
                ]
            )]

    def reset(self):
        self.selection_changed = False
        region = Region(self.X)
        self.update_region(region)
        # self.vs_rules_wgt.reset_widget()
        # self.es_rules_wgt.reset_widget()

    def update_region(self, region: Region):
        # TODO : compute score
        self.region = region
        self._update_title_txt()
        self.vs_rules_wgt.init_rules(region.rules, {}, region.mask)
        self.es_rules_wgt.init_rules(RuleSet(), {}, region.mask)
        self.update_selection(region.mask)
        if self.valid_selection:
            self.new_rules_defined(None, region.mask)

    def update_selection(self, selection_mask):
        self.selection_mask = selection_mask
        self.selection_changed = self.valid_selection
        X_rounded = self.X.loc[selection_mask].copy().apply(format_data)
        self.data_table.items = X_rounded.to_dict("records")
        self.update_selection_status()
        self.refresh_buttons()

    def update_selection_status(self):
        if self.valid_selection:
            selection_status_str_1 = f"{self.selection_mask.sum()} point selected"
            selection_status_str_2 = f"{100 * self.selection_mask.mean():.2f}% of the  dataset"
        else:
            selection_status_str_1 = f"0 point selected"
            selection_status_str_2 = f"0% of the  dataset"
        self.selection_status_str_1 = selection_status_str_1
        self.selection_status_str_2 = selection_status_str_2

    def update_X_exp(self, X_exp: pd.DataFrame):
        self.X_exp = X_exp
        self.es_rules_wgt.update_X(X_exp)

    def refresh_buttons(self):
        # data table
        self.widget[2].children[0].disabled = not self.valid_selection
        # skope_rule
        self.find_rules_btn.disabled = not self.valid_selection or not self.selection_changed
        # cancel
        self.cancel_btn.disabled = (self.vs_rules_wgt.rules_num == 0) and (self.region.num >= -1)
        # undo
        self.undo_btn.disabled = not (self.vs_rules_wgt.rules_num > 1)
        # validate rule
        self.validate_btn.disabled = not (self.vs_rules_wgt.rules_num > 0)

    @log_errors
    def compute_skope_rules(self, *args):
        self.selection_changed = False
        # compute skope rules
        skr_rules_list, skr_score_dict = skope_rules(self.selection_mask, self.X, self.variables)
        skr_score_dict['target_avg'] = self.y[self.selection_mask].mean()
        # init vs rules widget
        self.vs_rules_wgt.init_rules(skr_rules_list, skr_score_dict, self.selection_mask)
        # update VS and ES HDE
        self.update_callback(selection_mask=self.selection_mask, rules_mask=skr_rules_list.get_matching_mask(self.X))

        es_skr_rules_list, es_skr_score_dict = skope_rules(self.selection_mask, self.X_exp, self.variables)
        es_skr_score_dict['target_avg'] = self.y[self.selection_mask].mean()
        self.es_rules_wgt.init_rules(es_skr_rules_list, es_skr_score_dict, self.selection_mask)
        self.refresh_buttons()
        stats_logger.log('find_rules', skr_score_dict)

    @log_errors
    def undo_rules(self, *args):
        if self.vs_rules_wgt.rules_num > 0:
            self.vs_rules_wgt.undo()
        else:
            self.es_rules_wgt.undo()
        self.refresh_buttons()

    @log_errors
    def cancel_edit(self, *args):
        self.update_region(Region(self.X))
        self.refresh_buttons()

    @log_errors
    def new_rules_defined(self, rules_widget: RulesWidget | None, df_mask: pd.Series):
        stats_logger.log('rule_changed')
        """
        Called by a RulesWidget Skope rule creation or when the user wants new rules to be plotted
        The function asks the HDEs to display the rules result
        """
        # We sent to the proper HDE the rules_indexes to render :
        self.update_callback(selection_mask=self.selection_mask, rules_mask=df_mask)

        # sync selection between rules_widgets
        if rules_widget != self.vs_rules_wgt:
            self.vs_rules_wgt.update_from_mask(df_mask, RuleSet(), sync=False)
        if rules_widget != self.es_rules_wgt:
            self.es_rules_wgt.update_from_mask(df_mask, RuleSet(), sync=False)

        self.refresh_buttons()

    @log_errors
    def validate_rules(self, *args):
        stats_logger.log('validate_rules')
        # get rule set and check validity
        rules_set = self.vs_rules_wgt.current_rules_list
        if len(rules_set) == 0:
            stats_logger.log('validate_rules', info={'error': 'invalid rules'})
            self.vs_rules_wgt.show_msg("No rules found on Value space cannot validate region", "red--text")
            return

        # we persist the rule set in the region
        self.region.update_rule_set(rules_set)
        # we ship the region to GUI to synchronize other tab
        self.validate_rules_callback(self.region)
        # we reset the tab
        self.reset()



from functools import partial

import ipyvuetify as v
import pandas as pd
from auto_cluster import AutoCluster
from antakia_core.compute.skope_rule.skope_rule import skope_rules
from antakia_core.data_handler.region import RegionSet

from antakia import config
from antakia.gui.graphical_elements.color_table import ColorTable
from antakia.gui.high_dim_exp.projected_values_selector import ProjectedValuesSelector
from antakia.gui.helpers.progress_bar import MultiStepProgressBar
from antakia.utils.stats import stats_logger, log_errors


class Tab2:
    region_headers = [
        {
            "text": column,
            "sortable": False,
            "value": column,
        }
        for column in ['Region', 'Rules', 'Average', 'Points', '% dataset', 'Sub-model']
    ]

    def __init__(
        self,
        variables,
        X: pd.DataFrame,
        vs_pvs: ProjectedValuesSelector,
        es_pvs: ProjectedValuesSelector,
        region_set: RegionSet,
        edit_callback: callable,
        update_callback: callable,
        substitute_callback: callable
    ):
        self.X = X
        self.vs_pvs = vs_pvs
        self.es_pvs = es_pvs
        self.variables = variables
        self.region_set = region_set
        self.edit_callback = partial(edit_callback, self)
        self.update_callback = partial(update_callback, self, self.region_set)
        self.substitute_callback = partial(substitute_callback, self)
        self.auto_cluster_running = False

        self._build_widget()

    def _build_widget(self):
        self.region_table_wgt = ColorTable(  # 44001
            headers=self.region_headers,
            items=[],
        )
        self.stats_wgt = v.Html(  # 44002
            tag="p",
            class_="ml-2 mb-2",
            children=["0 region, 0% of the dataset"],
        )
        self.substitute_btn = v.Btn(  # 4401000
            v_on='tooltip.on',
            class_="ml-3 mt-8 green white--text",
            children=[
                v.Icon(
                    class_="mr-2",
                    children=[
                        "mdi-swap-horizontal-circle-outline"
                    ],
                ),
                "Substitute",
            ],
        )
        self.divide_btn = v.Btn(  # 4401100
            v_on='tooltip.on',
            class_="ml-3 mt-8",
            children=[
                v.Icon(
                    class_="mr-2",
                    children=[
                        "mdi-content-cut"
                    ],
                ),
                "Divide",
            ],
        )
        self.edit_btn = v.Btn(  # 4401100
            v_on='tooltip.on',
            class_="ml-3 mt-3",
            children=[
                v.Icon(
                    class_="mr-2",
                    children=[
                        "mdi-pencil"
                    ],
                ),
                "Edit",
            ],
        )
        self.merge_btn = v.Btn(  # 4401200
            v_on='tooltip.on',
            class_="ml-3 mt-3",
            children=[
                v.Icon(
                    class_="mr-2",
                    children=[
                        "mdi-table-merge-cells"
                    ],
                ),
                "Merge",
            ],
        )
        self.delete_btn = v.Btn(  # 4401300
            v_on='tooltip.on',
            class_="ml-3 mt-3",
            children=[
                v.Icon(
                    class_="mr-2",
                    children=[
                        "mdi-delete"
                    ],
                ),
                "Delete",
            ],
        )
        self.auto_cluster_btn = v.Btn(  # 4402000
            v_on='tooltip.on',
            class_="mt-8 primary",
            children=[
                v.Icon(  # 44020000
                    class_="mr-2",
                    children=[
                        "mdi-auto-fix"
                    ],
                ),
                "Auto-clustering",
            ],
        )
        self.cluster_num_wgt = v.Slider(  # 4402100
            v_on='tooltip.on',
            class_="mt-10 mx-2",
            v_model=6,
            min=2,
            max=12,
            thumb_color='blue',  # marker color
            step=1,
            thumb_label="always"
        )
        self.auto_cluster_checkbox = v.Checkbox(  # 440211
            class_="px-3",
            v_model=True,
            label="Automatic number of clusters"
        )
        self.auto_cluster_progress = v.ProgressLinear(  # 440212
            style_="width: 100%",
            class_="px-3",
            v_model=0,
            color="primary",
            height="15",
        )

        self.widget = [
            v.Row(  # 440
                class_="d-flex flex-row",
                children=[
                    v.Col(  # v.Sheet Col 1 # 4400
                        class_="col-8",
                        children=[
                            v.Html(  # 44000
                                tag="h3",
                                class_="ml-2",
                                children=["Regions :"],
                            ),
                            self.region_table_wgt,
                            self.stats_wgt,
                        ]
                    ),  # End Col 1
                    v.Col(  # v.Sheet Col 2 = buttons #4401
                        class_="col-2",
                        style_="size: 50%",
                        children=[
                            v.Row(  # 44010
                                class_="flex-column",
                                children=[
                                    v.Tooltip(  # 440100
                                        bottom=True,
                                        v_slots=[
                                            {
                                                'name': 'activator',
                                                'variable': 'tooltip',
                                                'children':
                                                    self.substitute_btn
                                            }
                                        ],
                                        children=[
                                            'Find an explicable surrogate model on this region']
                                    )
                                ]
                            ),
                            v.Row(  # 44011
                                class_="flex-column",
                                children=[
                                    v.Tooltip(  # 440110
                                        bottom=True,
                                        v_slots=[
                                            {
                                                'name': 'activator',
                                                'variable': 'tooltip',
                                                'children':
                                                    self.divide_btn
                                            }
                                        ],
                                        children=[
                                            'Divide a region into sub-regions']
                                    )
                                ]
                            ),
                            v.Row(  # 44011
                                class_="flex-column",
                                children=[
                                    v.Tooltip(  # 440110
                                        bottom=True,
                                        v_slots=[
                                            {
                                                'name': 'activator',
                                                'variable': 'tooltip',
                                                'children':
                                                    self.edit_btn
                                            }
                                        ],
                                        children=[
                                            'Edit region\'s rules']
                                    )
                                ]
                            ),
                            v.Row(  # 44012
                                class_="flex-column",
                                children=[
                                    v.Tooltip(  # 440120
                                        bottom=True,
                                        v_slots=[
                                            {
                                                'name': 'activator',
                                                'variable': 'tooltip',
                                                'children':
                                                    self.merge_btn
                                            }
                                        ],
                                        children=[
                                            'Merge regions']
                                    )
                                ]
                            ),
                            v.Row(  # 44013
                                class_="flex-column",
                                children=[
                                    v.Tooltip(  # 440130
                                        bottom=True,
                                        v_slots=[
                                            {
                                                'name': 'activator',
                                                'variable': 'tooltip',
                                                'children':
                                                    self.delete_btn
                                            }
                                        ],
                                        children=[
                                            'Delete region']
                                    )
                                ]
                            ),
                        ]  # End v.Sheet Col 2 children
                    ),  # End v.Sheet Col 2 = buttons
                    v.Col(  # v.Sheet Col 3 # 4402
                        class_="col-2 px-6",
                        style_="size: 50%",
                        children=[
                            v.Row(  # 44020
                                class_="flex-column",
                                children=[
                                    v.Tooltip(  # 440200
                                        bottom=True,
                                        v_slots=[
                                            {
                                                'name': 'activator',
                                                'variable': 'tooltip',
                                                'children':
                                                    self.auto_cluster_btn
                                            }
                                        ],
                                        children=[
                                            'Find homogeneous regions in both spaces']
                                    )
                                ]
                            ),
                            v.Row(  # 44021
                                class_="flex-column",
                                children=[
                                    v.Tooltip(  # 440210
                                        bottom=True,
                                        v_slots=[
                                            {
                                                'name': 'activator',
                                                'variable': 'tooltip',
                                                'children':
                                                    self.cluster_num_wgt,
                                            }
                                        ],
                                        children=['Number of clusters you expect to find']
                                    ),
                                    self.auto_cluster_checkbox,
                                    self.auto_cluster_progress
                                ]
                            ),
                        ]  # End v.Sheet Col 3 children
                    )  # End v.Sheet Col 3
                ]  # End v.Sheet children
            ),  # End v.Sheet
        ]

        self.wire()
        self.update_region_table()

    def wire(self):
        self.region_table_wgt.set_callback(self.region_selected)
        self.substitute_btn.on_event("click", self.substitute_clicked)
        self.edit_btn.on_event("click", self.edit_region_clicked)
        self.divide_btn.on_event("click", self.divide_region_clicked)
        self.merge_btn.on_event("click", self.merge_region_clicked)
        self.delete_btn.on_event("click", self.delete_region_clicked)
        self.auto_cluster_btn.on_event("click", self.auto_cluster_clicked)
        self.auto_cluster_checkbox.v_model = True
        self.auto_cluster_checkbox.on_event("change", self.checkbox_auto_cluster_clicked)
        self.cluster_num_wgt.on_event("change", self.num_cluster_changed)

    @property
    def selected_regions(self):
        return self.region_table_wgt.selected

    @selected_regions.setter
    def selected_regions(self, value):
        self.region_table_wgt.selected = value
        self.update_btns()

    def update_stats(self):
        region_stats = self.region_set.stats()
        str_stats = [
            f"{region_stats['regions']} {'regions' if region_stats['regions'] > 1 else 'region'}",
            f"{region_stats['points']} points",
            f"{region_stats['coverage']}% of the dataset",
            f"{region_stats['delta_score']:.2f} subst score"
        ]
        self.stats_wgt.children = [', '.join(str_stats)]

    def update_region_table(self):
        """
        Called to empty / fill the RegionDataTable and refresh plots
        """
        self.region_set.sort(by='size', ascending=False)
        temp_items = self.region_set.to_dict()

        # We populate the ColorTable :
        self.region_table_wgt.items = temp_items

        self.update_stats()
        self.update_btns()

    @log_errors
    def checkbox_auto_cluster_clicked(self, widget, event, data):
        """
        Called when the user clicks on the 'auto-cluster' checkbox
        """
        self.update_btns()

    @log_errors
    def auto_cluster_clicked(self, *args):
        """
        Called when the user clicks on the 'auto-cluster' button
        """
        # We disable the AC button. Il will be re-enabled when the AC progress is 100%
        self.auto_cluster_running = True
        if self.region_set.stats()["coverage"] > 80:
            # UI rules :
            # region_set coverage is > 80% : we need to clear it to do another auto-cluster
            self.region_set.clear_unvalidated()

        # We assemble indices ot all existing regions :
        region_set_mask = self.region_set.mask
        not_rules_indexes_list = ~region_set_mask
        # We call the auto_cluster with remaining X and explained(X) :
        cluster_num = self._get_cluster_num()
        stats_logger.log('auto_cluster', {
            'cluster_num': cluster_num,
            'vs_proj': str(self.vs_pvs.current_proj),
            'es_proj': str(self.es_pvs.current_proj)
        })

        self._compute_auto_cluster(not_rules_indexes_list, cluster_num)
        self.update_region_table()
        self.update_callback()
        # We re-enable the button
        self.auto_cluster_running = False
        self.update_btns()

    def num_cluster_changed(self, *args):
        """
        Called when the user changes the number of clusters
        """
        self.update_btns()

    def _get_cluster_num(self):
        if self.auto_cluster_checkbox.v_model:
            cluster_num = "auto"
        else:
            cluster_num = self.cluster_num_wgt.v_model - len(self.region_set)
            if cluster_num <= 2:
                cluster_num = 2
        return cluster_num

    def _compute_auto_cluster(self, not_rules_indexes_list, cluster_num='auto'):
        if len(not_rules_indexes_list) > config.ATK_MIN_POINTS_NUMBER:
            vs_compute = int(not self.vs_pvs.is_computed(dim=3))
            es_compute = int(not self.es_pvs.is_computed(dim=3))
            steps = 1 + vs_compute + es_compute

            progress_bar = MultiStepProgressBar(self.auto_cluster_progress, steps=steps)
            step = 1
            vs_proj_3d_df = self.vs_pvs.get_current_X_proj(
                3,
                progress_callback=progress_bar.get_update(step)
            )

            step += vs_compute
            es_proj_3d_df = self.es_pvs.get_current_X_proj(
                3,
                progress_callback=progress_bar.get_update(step)
            )

            step += es_compute
            ac = AutoCluster(self.X, progress_bar.get_update(step))

            found_regions = ac.compute(
                vs_proj_3d_df.loc[not_rules_indexes_list],
                es_proj_3d_df.loc[not_rules_indexes_list],
                # We send 'auto' or we read the number of clusters from the Slider
                cluster_num,
            )  # type: ignore
            self.region_set.extend(found_regions)
            progress_bar.set_progress(100)
        else:
            print('not enough points to cluster')

    def update_btns(self, current_operation=None):
        selected_region_nums = [x['Region'] for x in self.selected_regions]
        if current_operation:
            if current_operation['type'] == 'select':
                selected_region_nums.append(current_operation['region_num'])
            elif current_operation['type'] == 'unselect':
                selected_region_nums.remove(current_operation['region_num'])
        num_selected_regions = len(selected_region_nums)
        if num_selected_regions:
            first_region = self.region_set.get(selected_region_nums[0])
            enable_div = (num_selected_regions == 1) and bool(first_region.num_points() >= config.ATK_MIN_POINTS_NUMBER)
        else:
            enable_div = False

        # substitute
        self.substitute_btn.disabled = num_selected_regions != 1

        # edit
        self.edit_btn.disabled = num_selected_regions != 1

        # divide
        self.divide_btn.disabled = not enable_div

        # merge
        enable_merge = (num_selected_regions > 1)
        self.merge_btn.disabled = not enable_merge

        # delete
        self.delete_btn.disabled = num_selected_regions == 0

        # auto_cluster
        self.auto_cluster_btn.disabled = self.auto_cluster_running
        self.cluster_num_wgt.disabled = bool(self.auto_cluster_checkbox.v_model)

    def region_selected(self, data):
        operation = {
            'type': 'select' if data['value'] else 'unselect',
            'region_num': data['item']['Region']
        }
        self.update_btns(operation)

    def clear_selected_regions(self):
        self.selected_regions = []
        self.update_btns(None)

    @log_errors
    def edit_region_clicked(self, *args):
        """
        Called when the user clicks on the 'divide' (region) button
        """
        stats_logger.log('edit_region')
        # we recover the region to sudivide
        region = self.region_set.get(self.selected_regions[0]['Region'])
        self.edit_callback(region)

    @log_errors
    def divide_region_clicked(self, *args):
        """
        Called when the user clicks on the 'divide' (region) button
        """
        stats_logger.log('divide_region')
        # we recover the region to sudivide
        region = self.region_set.get(self.selected_regions[0]['Region'])
        if region.num_points() > config.ATK_MIN_POINTS_NUMBER:
            # Then we delete the region in self.region_set
            self.region_set.remove(region.num)
            # we compute the subregions and add them to the region set
            cluster_num = self._get_cluster_num()
            self._compute_auto_cluster(region.mask, cluster_num)
        # There is no more selected region
        self.clear_selected_regions()
        self.update_region_table()
        self.update_callback()

    @log_errors
    def merge_region_clicked(self, *args):
        """
        Called when the user clicks on the 'merge' (regions) button
        """

        selected_regions = [self.region_set.get(r['Region']) for r in self.selected_regions]
        stats_logger.log('merge_region', {'num_regions': len(selected_regions)})
        mask = None
        for region in selected_regions:
            if mask is None:
                mask = region.mask
            else:
                mask |= region.mask

        # compute skope rules
        skr_rules_list, _ = skope_rules(mask, self.X, self.variables)

        # delete regions
        for region in selected_regions:
            self.region_set.remove(region.num)
        # add new region
        if len(skr_rules_list) > 0:
            r = self.region_set.add_region(rules=skr_rules_list)
        else:
            r = self.region_set.add_region(mask=mask)
        self.selected_regions = [{'Region': r.num}]
        self.update_region_table()
        self.update_callback()

    @log_errors
    def delete_region_clicked(self, *args):
        """
        Called when the user clicks on the 'delete' (region) button
        """
        stats_logger.log('merge_region', {'num_regions': len(self.selected_regions)})
        for selected_region in self.selected_regions:
            region = self.region_set.get(selected_region['Region'])
            # Then we delete the regions in self.region_set
            self.region_set.remove(region.num)

        # There is no more selected region
        self.clear_selected_regions()
        self.update_region_table()
        self.update_callback()

    @log_errors
    def substitute_clicked(self, widget, event, data):
        stats_logger.log('substitute_region')
        region = self.region_set.get(self.selected_regions[0]['Region'])
        self.substitute_callback(region)



import pandas as pd

import ipyvuetify as v
from plotly.graph_objects import FigureWidget, Histogram

from antakia_core.data_handler.rules import Rule, RuleSet

from antakia_core.utils.utils import compute_step, get_mask_comparison_color
from antakia_core.utils.variable import DataVariables

from antakia.utils.stats import log_errors


class RuleWidget:
    """
    A piece of UI handled by a RulesExplorer
    It allows the user to modify one rule

    Attributes :
    """

    def __init__(self, rule: Rule, X: pd.DataFrame, y, values_space: bool, init_selection_mask: pd.Series,
                 init_rules_mask: pd.Series,
                 rule_updated: callable):
        '''

        Parameters
        ----------
        rule : the rule to be dsplayed
        X : training dataset
        y : target variable
        values_space : bool is value space ?
        init_selection_mask : reference selection mask
        init_rules_mask : reference rules mask
        rule_updated : callable called on update
        '''
        self.rule: Rule = rule
        self.X: pd.DataFrame = X
        self.X_col = X.loc[:, rule.variable.column_name]
        self.values_space: bool = values_space
        self.rule_updated: callable = rule_updated
        self.display_sliders: bool = self.values_space  # enable rule edit
        self.widget = None
        self.init_mask = init_selection_mask
        self.rule_mask = init_rules_mask

        self._build_widget()

    # --------------- build widget ------------------- #
    def _build_widget(self):
        """
        build the widget
        Returns
        -------

        """
        # build slider
        self.select_widget = self._get_select_widget()
        # build figure
        self._build_figure()

        title = self._get_panel_title()

        # root_widget is an ExpansionPanel
        self.widget: v.ExpansionPanel = v.ExpansionPanel(
            children=[
                v.ExpansionPanelHeader(
                    class_="blue lighten-4",
                    children=[title]
                ),
                v.ExpansionPanelContent(
                    children=[
                        v.Col(
                            children=[v.Spacer(), self.select_widget],
                        ),
                        self.figure,
                    ]
                ),
            ]
        )

        # The variable name bg (ExpansionPanelHeader) is light blue
        # get_widget(self.root_widget, "0").class_ = "blue lighten-4"

    def _build_figure(self):
        """
        draw the histograms
        Returns
        -------

        """
        base_args = {
            'bingroup': 1,
            'nbinsx': 50,
        }
        mask_color, colors_info = get_mask_comparison_color(self.rule_mask, self.init_mask)
        h = []
        for name, color in colors_info.items():
            h.append(Histogram(
                name=name,
                x=self.X_col[mask_color == color],
                marker_color=color,
                **base_args
            ))
        self.figure = FigureWidget(
            data=h
        )
        self.figure.update_layout(
            barmode="stack",
            bargap=0.1,
            # width=600,
            showlegend=False,
            margin={'l': 0, 'r': 0, 't': 0, 'b': 0},
            height=200,
        )

    def _get_panel_title(self):
        """
        compute and display accordion title
        Returns
        -------

        """
        if self.rule.is_categorical_rule:
            title = f"{self.rule.variable.display_name} possible values :"
        elif self.rule.rule_type == 1:  # var < max
            title = f"{self.rule.variable.display_name} lesser than {'or equal to ' if self.rule.includes_max else ''}:"
        elif self.rule.rule_type == 2:  # var > min
            title = f"{self.rule.variable.display_name} greater than {'or equal to ' if self.rule.includes_min else ''}:"
        elif self.rule.rule_type == 3:
            title = f"{self.rule.variable.display_name} inside the interval:"
        else:
            title = f"{self.rule.variable.display_name} outside the interval:"
        return title

    def _get_select_widget(self):
        """
        builds the widget to edit the rule
        Returns
        -------

        """
        if self.display_sliders:

            if self.rule.is_categorical_rule:
                return v.Select(
                    label=self.rule.variable.display_name,
                    items=self.X[self.rule.variable.column_name].unique().tolist(),
                    style_="width: 150px",
                    multiple=True,
                )
            min_ = float(self.X[self.rule.variable.column_name].min())
            max_ = float(self.X[self.rule.variable.column_name].max())
            min_, max_, step = compute_step(min_, max_)
            min_ = min_ - step
            slider_args = {
                'min': min_,
                'max': max_,
                'step': step,
                'thumb_label': "always",
                'thumb_size': 30,
                'thumb_color': 'blue'
            }
            if self.rule.rule_type == 1:  # var < max
                slider_class = v.Slider
                slider_args['color'] = 'green'
                slider_args['track_color'] = 'red'
            elif self.rule.rule_type == 2:  # var > min
                slider_class = v.Slider
                slider_args['color'] = 'red'
                slider_args['track_color'] = 'green'
            elif self.rule.is_inner_interval_rule:
                slider_class = v.RangeSlider
                slider_args['color'] = 'green'
                slider_args['track_color'] = 'red'
            else:
                slider_class = v.RangeSlider
                slider_args['color'] = 'red'
                slider_args['track_color'] = 'green'

            slider = slider_class(**slider_args)

            slider.v_model = self._get_select_widget_values()
            slider.on_event("change", self._widget_value_changed)
            return slider
        else:
            return v.Col()

    def _get_select_widget_values(self):
        """
        sets the selection values
        Returns
        -------

        """
        if self.rule.is_categorical_rule:
            return self.rule.cat_values
        elif self.rule.rule_type == 1:  # var < max
            return [self.rule.max],
        elif self.rule.rule_type == 2:  # var > min
            return [self.rule.min]
        else:
            return [self.rule.min, self.rule.max]

    # --------------- callbacks ------------------- #
    @log_errors
    def _widget_value_changed(self, widget, event, data):
        """
        callback called when the user edits a value (called by the widget)

        should fire rule updated callback

        Parameters
        ----------
        widget
        event
        data : new value

        Returns
        -------

        """
        cat_values = None
        if self.rule.is_categorical_rule:
            cat_values = data
            min_ = max_ = None
        else:
            if isinstance(data, list):  # Interval rule
                min_ = data[0]
                max_ = data[1]
            elif self.rule.rule_type == 1:  # var < max
                min_ = None
                max_ = data
            else:
                min_ = data
                max_ = None
        new_rule = Rule(
            self.rule.variable,
            min_,
            self.rule.includes_min,
            max_,
            self.rule.includes_max,
            cat_values
        )
        self.rule = new_rule
        self.rule_updated(new_rule)

    def update(self, new_rules_mask: pd.Series = None, rule: Rule = None):
        """ 
            used to update the display (sliders and histogram) to match the new rule
            (called from outside th object to synchronize it)
        """

        if rule is not None:
            self.rule = rule
        if new_rules_mask is not None:
            self.rule_mask = new_rules_mask
        # We update the selects
        if self.display_sliders:
            self._get_select_widget_values()
        mask_color, colors_info = get_mask_comparison_color(self.rule_mask, self.init_mask)
        with self.figure.batch_update():
            for i, color in enumerate(colors_info.values()):
                self.figure.data[i].x = self.X_col[mask_color == color]


class RulesWidget:
    """
    A RulesWidget is a piece of GUI that allows the user to refine a set of rules.
    The user can use the slider to change the rules.
    There are 2 RW : VS and ES slides

    rules_db : a dict of list :
        [[rules_list, scores]]], so that at iteration i we have :
        [i][0] : the list of rules
        [i][1] : the scores as a dict {"precision": ?, "recall": ?, "f1": ?}
    X : pd.DataFrame, values or explanations Dataframe depending on the context
    variables : list of Variable
    rules_mask : list of indexes of the GUI selection (from X.index)
    is_value_space : bool
    is_disabled : bool
    rules_mask : Dataframe mask of the rules
    rules_updated : callable of the GUI parent
    root_wiget : its widget representation
    """

    def __init__(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        variables: DataVariables,
        values_space: bool,
        new_rules_defined: callable = None,
    ):
        """
        widget to manage rule edition and display
        Parameters
        ----------
        X: pd.Dataframe reference dataset
        y: pd.Series test dataset
        variables
        values_space: is the widget on the value space
        new_rules_defined: callback on rule update
        """
        self.X = X
        self.y = y
        self.variables: DataVariables = variables
        self.init_selection_mask = None
        self.is_value_space = values_space
        self.new_rules_defined = new_rules_defined

        self.rules_db: list[tuple[RuleSet, dict]] = []
        self.rule_widget_list: list[RuleWidget] = []

        # At startup, we are disabled :
        self.is_disabled = True

        self._build_widget()

    # ---------------- widget management -------------------- #

    def _build_widget(self):
        self.title = "not initialized"
        self.stats = "not initialized"
        self.rules = 'N/A'
        self.region_stat_card = v.Col(  # 43100 / 0
            children=[
                v.Row(  # 431000 / 00
                    children=[
                        v.Icon(children=["mdi-target"]),  #
                        v.Html(class_="ml-3", tag="h2",
                               children=[self.title]),
                    ]
                ),
                v.Html(  # 431001 / 01
                    class_="ml-7",
                    tag="li",
                    children=[
                        self.stats
                    ]
                ),
                v.Html(  # 431002 / 02
                    class_="ml-7",
                    tag="li",
                    children=[self.rules]
                ),
            ]
        )
        self.rules_widgets = v.ExpansionPanels(  # Holds VS RuleWidgets  # 43101 / 1
            style_="max-width: 95%",
            children=[rw.widget for rw in self.rule_widget_list]
        )
        self.widget = v.Col(  # placeholder for the VS RulesWidget (RsW) # 4310
            children=[self.region_stat_card, self.rules_widgets],
        )
        self.refresh_widget()
        self.disable()

    def set_title(self, title: str, css: str):
        self.title = title
        title_wgt = self.region_stat_card.children[0].children[1]
        title_wgt.children = [self.title]
        title_wgt.class_ = css

    def set_stats(self, stats: str, css: str):
        self.stats = stats

        stats_wgt = self.region_stat_card.children[1]
        stats_wgt.children = [self.stats]
        stats_wgt.class_ = css

    def set_rules(self, rules: str, css: str):
        self.rules = rules

        rules_wgt = self.region_stat_card.children[2]
        rules_wgt.children = [self.rules]
        rules_wgt.class_ = css

    def set_rules_widgets(self, rule_widget_list: list[RuleWidget]):
        self.widget.children[1].children = [rule_widget.widget for rule_widget in rule_widget_list]

    def _create_rule_widgets(self):
        """
        creates all rule widgets
                Called by self.init_rules
                If init_rules_mask is None, we clear all our RuleWidgets

        Parameters
        ----------
        init_mask : reference_mask

        Returns
        -------

        """
        init_rule_mask = self.current_rules_list.get_matching_mask(self.X)
        # We set new RuleWidget list and put it in our ExpansionPanels children
        if len(self.current_rules_list) == 0:
            self.rule_widget_list = []
        else:
            self.rule_widget_list = [
                RuleWidget(rule, self.X, self.y, self.is_value_space,
                           self.init_selection_mask, init_rule_mask,
                           self.update_rule)
                for rule in self.current_rules_list.rules.values()
            ]
        self.set_rules_widgets(self.rule_widget_list)

    # ------------- widget macro method ------------- #
    def disable(self, is_disabled: bool = True):
        """
        disables (True) or enables (False) the widget
        Disabled : card in grey with dummy text + no ExpansionPanels
        Enabled : card in light blue / ready to display RuleWidgets
        """

        self.is_disabled = is_disabled

        self.refresh_widget()

    def refresh_widget(self):
        """
        Sets macro widget info and Datatable: scores and rule details int the DataTable
        Returns
        -------

        """
        # We set the title
        self._show_title()

        # We set the scores
        self._show_score()

        # We set the rules
        self._show_rules()

        self._show_rules_wgt()

    def _show_title(self):
        if len(self.current_rules_list) == 0 or self.is_disabled:
            title = f"No rule to display for the {'VS' if self.is_value_space else 'ES'} space"
            css = "ml-3 grey--text" if self.is_disabled else "ml-3"
        else:
            title = f"Rule(s) applied to the {'values' if self.is_value_space else 'explanations'} space"
            css = "ml-3"
        self.set_title(title, css)

    def _show_score(self):
        """
        show rule score
        Returns
        -------

        """
        if len(self.current_scores_dict) == 0 or self.is_disabled:
            scores_txt = "Precision = n/a, recall = n/a, f1_score = n/a"
            css = "ml-7 grey--text"
        elif self.current_scores_dict['precision'] == -1:
            scores_txt = "No point of the dataset matches the new rules"
            css = "ml-7 red--text"
        else:
            precision, recall, f1, target_avg = (
                self.current_scores_dict['precision'], self.current_scores_dict['recall'],
                self.current_scores_dict['f1'], self.current_scores_dict['target_avg'],
            )
            scores_txt = (f"Precision : {precision:.2f}, recall :{recall:.2f} ," +
                          f" f1_score : {f1:.2f}, target_avg : {target_avg:.2f}")
            css = "ml-7 black--text"
        self.set_stats(scores_txt, css)

    def _show_rules(self):
        """
        show rules as text
        Returns
        -------

        """
        if (
            len(self.current_rules_list) == 0
            or self.is_disabled
        ):
            rules_txt = "N/A"
            css = "ml-7 grey--text"
        else:
            rules_txt = repr(self.current_rules_list)
            css = "ml-7 blue--text"
        self.set_rules(rules_txt, css)

    def show_msg(self, msg: str, css: str = ""):
        """
        print a message for the user
        Parameters
        ----------
        msg : message to be printed
        css : css class to apply

        Returns
        -------

        """
        css = "ml-7 " + css
        self.set_rules(msg, css)

    def _show_rules_wgt(self):
        if self.is_disabled:
            self.set_rules_widgets([])
        else:
            self.set_rules_widgets(self.rule_widget_list)

    # ---------------- update logic -------------------------- #
    @property
    def space(self):
        return 'values' if self.is_value_space else 'explanations'

    def update_X(self, X):
        """
        change the reference dataset
        Parameters
        ----------
        X: pd.DataFrame

        Returns
        -------

        """
        if id(X) == id(self.X):
            return
        self.X = X
        self._create_rule_widgets()
        self.refresh_widget()

    def init_rules(self, rules_set: RuleSet, score_dict: dict, selection_mask: pd.Series):
        """
        initialize the widget with the rule list, the score dict (text displayed) and
        the reference_mask (selection_mask)
        """
        # we start wih a fresh widget
        self.reset_widget()

        self.init_selection_mask = selection_mask
        if len(rules_set) > 0:
            # self.show_msg("No rules found", "red--text")
            # return
            self.disable(False)
            # We populate the db
            self._put_in_db(rules_set, score_dict)

            # We create the RuleWidgets
            self._create_rule_widgets()
        # we display
        self.refresh_widget()

    def reset_widget(self):
        """
        a reseted widget is
        - disabled
        - history is erased
        - reference mask is None
        Returns
        -------

        """
        self.disable()
        self.rules_db = []
        self.rule_widget_list = []
        self.init_selection_mask = None
        self.refresh_widget()

    def update_rule(self, new_rule: Rule):
        """
        callback to synchronize rules in the widget
        called by the edition of a single rule
        """
        # We update the rule in the db
        new_rules_set = self.current_rules_list.copy()
        new_rules_set.add(new_rule)

        new_rules_mask = new_rules_set.get_matching_mask(self.X)
        self.update_from_mask(new_rules_mask, new_rules_set)

    def update_from_mask(self, new_rules_mask: pd.Series, new_rules_set: RuleSet, sync=True):
        """
        updates the widget with the new rule_mask and rule list - the reference_mask is kept for comparison
        Parameters
        ----------
        new_rules_mask : mask to display
        new_rules_list : rules to base widget on
        sync : whether or not to call the sync callback (self.new_rules_defined) to update the other rules widget

        Returns
        -------

        """
        if len(new_rules_set):
            # update rules
            try:
                precision = (new_rules_mask & self.init_selection_mask).sum() / new_rules_mask.sum()
                recall = (new_rules_mask & self.init_selection_mask).sum() / self.init_selection_mask.sum()
                f1 = 2 * (precision * recall) / (precision + recall)
            except ZeroDivisionError:
                precision = recall = f1 = -1
            target_avg = self.y[new_rules_mask].mean()

            new_score_dict = {"precision": precision, "recall": recall, "f1": f1, 'target_avg': target_avg}

            self._put_in_db(new_rules_set, new_score_dict)

        # We refresh the widget macro info
        self.refresh_widget()

        # We update each of our RuleWidgets
        for rw in self.rule_widget_list:
            rw.update(new_rules_mask)

        # We notify the GUI and tell there are new rules to draw
        if self.new_rules_defined is not None and sync:
            self.new_rules_defined(self, new_rules_mask)

    def undo(self):
        """
        Restore the previous rules
        """
        # We remove last rules item from the db:
        if len(self.rules_db) > 1:
            self.rules_db.pop(-1)

        # We compute again the rules mask
        rules_mask = self.current_rules_list.get_matching_mask(self.X)

        def find_rule(rule_widget: RuleWidget) -> Rule:
            var = rule_widget.rule.variable
            return self.current_rules_list.find_rule(var)

        # We update each of our RuleWidgets
        for rw in self.rule_widget_list:
            rw.update(rules_mask, find_rule(rw))

        # we refresh the widget macro info
        self.refresh_widget()
        # We notify the GUI and tell there are new rules to draw if necessary
        if self.new_rules_defined is not None:
            self.new_rules_defined(self, rules_mask)

    def _put_in_db(self, rule_set: RuleSet, score_dict: dict):
        """
        add the rule list to memory
        Parameters
        ----------
        rule_set
        score_dict

        Returns
        -------

        """
        self.rules_db.append((rule_set, score_dict))

    def _dump_db(self) -> str:
        """
        print db
        Returns
        -------

        """
        txt = ""
        for i in range(len(self.rules_db)):
            rules_list = self.rules_db[i][0]
            scores_dict = self.rules_db[i][1]

            txt += f"({i}) : {len(rules_list)} rules:"
            "\n    ".join(rules_list)
            txt += f"\n   scores = {scores_dict}\n"
        return txt

    @property
    def current_rules_list(self) -> RuleSet:
        """
        get the current rule list
        Returns
        -------

        """
        if len(self.rules_db) == 0:
            return RuleSet()
        else:
            return self.rules_db[-1][0]

    @property
    def current_scores_dict(self) -> dict:
        """
        get the current score dict
        Returns
        -------

        """
        if len(self.rules_db) == 0:
            return {}
        else:
            return self.rules_db[-1][1]

    @property
    def rules_num(self) -> int:
        """
        get the size of the db
        Returns
        -------

        """
        return len(self.rules_db)



from io import StringIO
import pandas as pd
import urllib.request
import requests
import json
from importlib.resources import files
import os

import logging
from antakia.utils.logging_utils import conf_logger

logger = logging.getLogger(__name__)
conf_logger(logger)

AVAILABLE_EXAMPLES = {
    "california_housing": [],
    "climate_change_survey": ["X_train", "X_test", "y_train", "y_test"],
    "wages": [],
    "titanic": ["train", "test"]
}

BRANCH = "dev"  # TODO change to "main" when merging


def _get_github_url(example: str, dataset_name: str = None) -> str:
    """
    Returns the Github url of the dataset from our examples/data folder

    Parameters
    ----------
    example : str
        The name of the example. Must be one of the following: "california_housing", "climate_change_survey", "wages"
    dataset_name : str (optional)
        The name of the dataset if needed. For example: "X_train", "X_test", "y_train", "y_test"
    
    Returns 
    -------
    str
        The Github url of the dataset requested of the specified example.
    """

    base_url = f"https://raw.githubusercontent.com/AI-vidence/antakia/{BRANCH}/examples/data/"

    if example not in AVAILABLE_EXAMPLES:
        raise ValueError(
            f"Example {example} not found, dataset should be one of {list(AVAILABLE_EXAMPLES.keys())}")
    if AVAILABLE_EXAMPLES[example] and dataset_name not in AVAILABLE_EXAMPLES[example]:
        raise ValueError(f"Dataset must be one of the following : {', '.join(AVAILABLE_EXAMPLES[example])}")
    if AVAILABLE_EXAMPLES[example]:
        url = base_url + example + "/" + dataset_name.lower() + ".csv"
        logger.debug(f"Dataset {example}/{dataset_name} download URL : {url}")
    else:
        url = base_url + example + ".csv"
        logger.debug(f"Dataset {example} download URL : {url}")
    return url


def _get_lfs_pointer(url: str) -> tuple[str, str, int]:
    """
    Reads LFS file metadata from its GitHub file, given its Github URL 

    Parameters
    ----------
    url : str
        The Github URL of the dataset
    
    Returns
    -------
    tuple[str, str, int]
        The LFS server URL, the SHA256 hash and the size of the file
    
    """
    path, _ = urllib.request.urlretrieve(url)
    f = open(path, 'r')  # TODO would be better not to use a file -> we should use requests.get(url).text
    meta = f.read().strip().split('\n')
    f.close()
    server_url = meta[0].split(' ')[1]
    sha256 = meta[1].split(':')[1]
    file_size = int(meta[2].split(' ')[1])
    logger.debug(f"LFS metadata : {server_url}, {sha256}, {file_size}")
    return server_url, sha256, file_size


def _get_download_url(oid: str, file_size: int) -> str:
    """
    Returns the download URL of a LFS Github file given its oid and size

    Parameters
    ----------
    oid : str
        The OID of the Github LFS file
    file_size : int
        The size of the file, in bytes
    
    Returns
    -------
    str
        The ream download URL of the file
    """

    url = "https://github.com/AI-vidence/antakia.git/info/lfs/objects/batch"
    payload = {
        'operation': 'download',
        'transfer': ['basic'],
        'objects': [
            {
                'oid': oid,
                'size': file_size
            }
        ]
    }
    headers = {'content-type': 'application/json', 'Accept': 'application/vnd.git-lfs+json'}

    # Note the json.dumps(payload) that serializes the dict to a string, otherwise requests doesn't understand nested dicts !
    r = requests.post(url, data=json.dumps(payload), headers=headers)

    dwl_url = r.json()["objects"][0]['actions']['download']['href']
    logger.debug(f"Download URL : {dwl_url}")
    return dwl_url


def _get_remote_dataset(example: str, dataset_name: str = None) -> pd.DataFrame:
    """
    Fetchs the dataset from github repository examples/data folder. Called by our examples notebooks.

    Parameters
    ----------
    example : str
        The name of the example. Must be one of the following: "california_housing", "climate_change_survey", "wages", "titanic"
    dataset_name : str (optional)
        The name of the dataset if needed. For example: "X_train", "X_test", "y_train", "y_test"

    Returns
    -------
    pd.DataFrame
        A DataFrame corresponding to the dataset requested of the specified example.
    """
    _, oid, file_size = _get_lfs_pointer(_get_github_url(example, dataset_name))
    url = _get_download_url(oid, file_size)
    r = requests.get(url)
    logger.debug(f"Fetched dataset file: {r.text}")
    return pd.read_csv(StringIO(r.text))


def fetch_dataset(example: str, dataset_name: str = None) -> pd.DataFrame | dict[str:pd.DataFrame]:
    """
    Fetchs the dataset from our examples/data folder. Called by our examples notebooks.

    Parameters
    ----------
    example : str
        The name of the example. Must be one of the following: "california_housing", "climate_change_survey", "wages", "titanic"
    dataset_name : str (optional)
        The name of the dataset if needed. For example: "X_train", "X_test", "y_train", "y_test"
    
    Returns 
    -------
    pd.DataFrame
        A DataFrame corresponding to the dataset requested of the specified example.
    """
    if dataset_name:
        file_name = f'{example}-{dataset_name}.csv'
    elif AVAILABLE_EXAMPLES[example]:
        datasets = {}
        for name in AVAILABLE_EXAMPLES[example]:
            datasets[name] = fetch_dataset(example, name)
        return datasets
    else:
        file_name = f'{example}.csv'
    example_folder = files("antakia").joinpath("assets/examples/")
    location = str(example_folder.joinpath(f"{file_name}"))
    if os.path.exists(location):
        dataset = pd.read_csv(location)
        if 'Unnamed: 0' == dataset.columns[0]:
            dataset = dataset.set_index(dataset.columns[0])
            dataset.index.name=None
        return dataset
    else:
        if not os.path.exists(example_folder):
            os.mkdir(example_folder)
        dataset = _get_remote_dataset(example, dataset_name)
        dataset.to_csv(location, index=False)
        if 'Unnamed: 0' == dataset.columns[0]:
            dataset = dataset.set_index(dataset.columns[0])
            dataset.index.name = None
        return dataset



from __future__ import annotations

import time

import numpy as np
import pandas as pd

import ipyvuetify as v
import IPython.display

from antakia_core.data_handler.region import ModelRegionSet, ModelRegion, Region

from antakia.gui.app_bar.color_switch import ColorSwitch
from antakia.gui.app_bar.dimension_switch import DimSwitch
from antakia.gui.splash_screen import SplashScreen
from antakia.gui.app_bar.top_bar import TopBar
from antakia.gui.app_bar.explanation_values import ExplanationValues
from antakia.gui.high_dim_exp.projected_value_bank import ProjectedValueBank
from antakia_core.explanation.explanation_method import ExplanationMethod
import antakia.config as config
from antakia_core.data_handler.rules import RuleSet

from antakia.gui.tabs.model_explorer import ModelExplorer
from antakia.gui.tabs.tab1 import Tab1
from antakia.gui.tabs.tab2 import Tab2
from antakia.gui.tabs.tab3 import Tab3
from antakia.gui.high_dim_exp.highdimexplorer import HighDimExplorer

from antakia.gui.helpers.metadata import metadata

import logging
from antakia.utils.logging_utils import conf_logger
from antakia_core.utils.utils import boolean_mask, ProblemCategory
from antakia_core.utils.variable import DataVariables

from antakia.utils.stats import stats_logger, log_errors

logger = logging.getLogger(__name__)
conf_logger(logger)


class GUI:
    """
    GUI class.

    The GUI guides the user through the AntakIA process.
    It stores Xs, Y and the model to explain.
    It displays a UI (app_graph) and creates various UI objects, in particular
    two HighDimExplorers resposnible to compute or project values in 2 spaces.

    The interface is built using ipyvuetify and plotly.
    It heavily relies on the IPyWidgets framework.

    Instance Attributes
    ---------------------
    X : Pandas DataFrame, the orignal dataset
    y : Pandas Series, target values
    model : a model
    X_exp : a Pandas DataFrame, containing imported explanations
    variables : a list of Variable
    selection_ids : a list of a pd.DataFrame indexes, corresponding to the current selected points
        IMPORTANT : a dataframe index may differ from the row number
    vs_hde, es_hde : HighDimExplorer for the VS and ES space
    vs_rules_wgt, es_rules_wgt : RulesWidget
    region_set : a list of Region,
        a region is a dict : {'num':int, 'rules': list of rules, 'indexes', 'model': str, 'score': str}
        if the list of rules is None, the region has been defined with auto-cluster
        num start at 1
    validated_rules_region, validated_region, validated_sub_model

    """

    def __init__(
            self,
            X: pd.DataFrame,
            y: pd.Series,
            model,
            variables: DataVariables,
            X_test: pd.DataFrame,
            y_test: pd.Series,
            X_exp: pd.DataFrame | None = None,
            score: callable | str = "mse",
            problem_category: ProblemCategory = ProblemCategory.regression
    ):
        metadata.start()
        self.tab = 1
        self.X = X
        self.X_test = X_test
        self.y = y
        self.y_test = y_test
        self._y_pred = None
        self.problem_category = problem_category
        self.model = model
        self.variables: DataVariables = variables
        self.score = score
        # Init value space widgets
        self.selection_mask = boolean_mask(X, True)

        self.pv_bank = ProjectedValueBank(y)
        self.region_set = ModelRegionSet(self.X, self.y, self.X_test, self.y_test, self.model, self.score)

        # star dialog
        self.topbar = TopBar()

        self.dimension_switch = DimSwitch(self.dimension_update_callback)
        self.color_switch = ColorSwitch(self.y, self.y_pred, self.color_update_callback)

        # first hde
        self.vs_hde = HighDimExplorer(
            self.pv_bank,
            self.selection_changed,
            'VS'
        )

        # init Explanation space
        # first explanation getter/compute
        self.exp_values = ExplanationValues(
            self.X,
            self.y,
            self.model,
            problem_category,
            self.explanation_changed_callback,
            self.disable_hde,
            X_exp
        )
        # then hde
        self.es_hde = HighDimExplorer(
            self.pv_bank,
            self.selection_changed,
            'ES'
        )

        # init tabs
        self.tab1 = Tab1(variables, self.new_rule_selected_callback, self.validate_rules_callback, self.X, X_exp,
                         self.y)

        self.tab2 = Tab2(
            variables,
            X,
            self.vs_hde.projected_value_selector,
            self.es_hde.projected_value_selector,
            self.region_set,
            self.edit_region_callback,
            self.update_region_callback,
            self.substitute_model_callback
        )

        self.tab3 = Tab3(X, problem_category, self.model_validation_callback)
        self.model_explorer = ModelExplorer(self.X)

        self._build_widget()
        self.splash = SplashScreen(X)

    def _build_widget(self):
        self.widget = v.Col(
            children=[
                self.topbar.widget,
                v.Row(  # Top buttons bar # 1
                    class_="mt-3 align-center",
                    children=[
                        v.Tooltip(  # 10
                            bottom=True,
                            v_slots=[
                                {
                                    'name': 'activator',
                                    'variable': 'tooltip',
                                    'children': self.dimension_switch.widget
                                }  # End v_slots dict
                            ],  # End v_slots list
                            children=['Change dimensions']
                        ),  # End v.Tooltip
                        self.color_switch.widget,
                        v.Col(  # 12
                            class_="ml-4 mr-4",
                            children=[self.exp_values.widget]
                        ),
                        v.Col(  # 13 VS proj Select
                            class_="ml-6 mr-6",
                            children=[self.vs_hde.projected_value_selector.widget]
                        ),
                        v.Col(  # 14 ES proj Select
                            class_="ml-6 mr-6",
                            children=[self.es_hde.projected_value_selector.widget]
                        ),

                    ]
                ),
                v.Row(  # The two HighDimExplorer # 2
                    class_="d-flex",
                    children=[
                        v.Col(  # VS HDE # 20
                            style_="width: 50%",
                            class_="d-flex flex-column justify-center",
                            children=[
                                v.Html(  # 200
                                    tag="h3",
                                    style_="align-self: center",
                                    class_="mb-3",
                                    children=["Values space"]
                                ),
                                self.vs_hde.figure.widget,
                            ],
                        ),
                        v.Col(  # ES HDE placeholder # 21
                            style_="width: 50%",
                            class_="d-flex flex-column justify-center",
                            children=[
                                v.Html(  # 210
                                    tag="h3",
                                    style_="align-self: center",
                                    class_="mb-3",
                                    children=["Explanations space"]
                                ),
                                self.es_hde.figure.widget
                            ],
                        ),
                    ],
                ),
                v.Divider(),  # 3
                v.Tabs(  # 4
                    v_model=0,  # default active tab
                    children=[
                                 v.Tab(children=["Selection"]),  # 40
                                 v.Tab(children=["Regions"]),  # 41
                                 v.Tab(children=["Substitution"]),  # 42
                             ]
                             +
                             [
                                 v.TabItem(  # Tab 1)
                                     class_="mt-2",
                                     children=self.tab1.widget
                                 ),
                                 v.TabItem(  # Tab 2) Regions #44
                                     children=self.tab2.widget
                                 ),  # End of v.TabItem #2
                                 v.TabItem(  # TabItem #3 Substitution #45
                                     children=self.tab3.widget
                                 )
                             ]
                )  # End of v.Tabs
            ]  # End v.Col children
        )  # End of v.Col

    def compute_base_values(self):
        # We trigger ES explain computation if needed :
        if not self.exp_values.has_user_exp:  # No imported explanation values
            exp_method = ExplanationMethod.explain_method_as_str(config.ATK_DEFAULT_EXPLANATION_METHOD)
            msg = f"Computing {exp_method} on {self.X.shape}"
        else:
            msg = f"Imported explained values {self.X.shape}"
        self.splash.set_exp_msg(msg)
        self.exp_values.initialize(self.splash.exp_progressbar)

        # We trigger VS proj computation :
        self.splash.set_proj_msg(f"{config.ATK_DEFAULT_PROJECTION} on {self.X.shape} 1/2")
        self.vs_hde.initialize(progress_callback=self.splash.proj_progressbar.get_update(1), X=self.X)

        # THen we trigger ES proj computation :
        self.splash.set_proj_msg(f"{config.ATK_DEFAULT_PROJECTION} on {self.X.shape} 2/2")
        self.es_hde.initialize(
            progress_callback=self.splash.proj_progressbar.get_update(2),
            X=self.exp_values.current_exp_df
        )
        self.tab1.update_X_exp(self.exp_values.current_exp_df)
        self.selection_changed(None, boolean_mask(self.X, True))

        self.select_tab(0)
        self.disable_hde()

        if metadata.counter == 10:
            self.topbar.open()

    @log_errors
    def initialize(self):
        """Displays the splash screen and updates it during the first computations."""

        # We add both widgets to the current notebook cell and hide them
        t = time.time()
        self.widget.hide()
        self.splash.widget.show()
        IPython.display.display(self.splash.widget, self.widget)

        self.compute_base_values()

        self.wire()

        self.splash.widget.hide()
        self.widget.show()
        # redraw figures once app is displayed to be able to autosize it
        self.vs_hde.figure.create_figure()
        self.es_hde.figure.create_figure()
        stats_logger.log('gui_init_end', {'load_time': time.time() - t})

    def wire(self):
        """
        wires the app_widget, and implements UI logic
        """

        # ================ Tab Selection ================

        # We wire the click event on 'Tabs'
        tabs = self.widget.children[4].children
        tabs[0].on_event("click", self.select_tab_front(1))
        tabs[1].on_event("click", self.select_tab_front(2))
        tabs[2].on_event("click", self.select_tab_front(3))

    # ==================== properties ==================== #

    @property
    def y_pred(self):
        if self._y_pred is None:
            pred = self.model.predict(self.X)
            if self.problem_category in [ProblemCategory.classification_with_proba]:
                pred = self.model.predict_proba(self.X)

            if len(pred.shape) > 1:
                if pred.shape[1] == 1:
                    pred = pred.squeeze()
                if pred.shape[1] == 2:
                    pred = np.array(pred)[:, 1]
                else:
                    pred = pred.argmax(axis=1)
            self._y_pred = pd.Series(pred, index=self.X.index)
        return self._y_pred

    # ==================== sync callbacks ==================== #

    @log_errors
    def explanation_changed_callback(self, current_exp_df: pd.DataFrame, progress_callback: callable = None):
        """
        on explanation change, synchronizes es_hde and tab1
        Parameters
        ----------
        current_exp_df
        progress_callback

        Returns
        -------

        """
        self.es_hde.update_X(current_exp_df, progress_callback)
        self.tab1.update_X_exp(current_exp_df)

    @log_errors
    def disable_hde(self, disable='auto'):
        if disable == 'auto':
            disable_proj = bool((self.tab == 0) and self.selection_mask.any() and not self.selection_mask.all())
            disable_figure = bool(self.tab > 1)
        else:
            disable_proj = disable
            disable_figure = disable
        self.vs_hde.disable(disable_figure, disable_proj)
        self.exp_values.disable_selection(disable_proj)
        self.es_hde.disable(disable_figure, disable_proj)

    @log_errors
    def selection_changed(self, caller: HighDimExplorer | None, new_selection_mask: pd.Series):
        """
        callback to synchronize both hdes and tab1
        Parameters
        ----------
        caller
        new_selection_mask

        Returns
        -------

        """
        """Called when the selection of one HighDimExplorer changes"""

        self.selection_mask = new_selection_mask

        # If new selection (empty or not) : if exists, we remove any 'pending rule'
        self.disable_hde()
        if new_selection_mask.all():
            # Selection is empty
            self.select_tab(0)
        else:
            stats_logger.log('selection_gui',
                             {'exp_method': self.exp_values.current_exp,
                              'vs_proj': str(self.vs_hde.projected_value_selector.current_proj),
                              'es_proj': str(self.es_hde.projected_value_selector.current_proj)})

        # We synchronize selection between the two HighDimExplorers
        if caller is None:
            self.es_hde.set_selection(self.selection_mask)
            self.vs_hde.set_selection(self.selection_mask)
        else:
            other_hde = self.es_hde if caller == self.vs_hde.figure else self.vs_hde
            other_hde.set_selection(self.selection_mask)

        # update tab1
        self.tab1.update_selection(self.selection_mask)

    # ==================== top bar ==================== #

    def dimension_update_callback(self, caller, dim):
        self.vs_hde.set_dim(dim)
        self.es_hde.set_dim(dim)
        self.disable_hde()

    @log_errors
    def color_update_callback(self, caller, color):
        """
        Called with the user clicks on the colorChoiceBtnToggle
        Allows change the color of the dots
        """
        self.vs_hde.figure.set_color(color, 0)
        self.es_hde.figure.set_color(color, 0)
        self.select_tab(0)

    # ==================== TAB handling ==================== #

    def select_tab_front(self, tab):
        @log_errors
        def call_fct(*args):
            stats_logger.log('tab_selected', {'tab': tab})
            self.select_tab(tab, front=True)

        return call_fct

    def select_tab(self, tab, front=False):
        print('select tab', tab, front)
        if tab == 1 and (not self.selection_mask.any() or self.selection_mask.all()):
            return self.select_tab(0)
        if tab == 1:
            self.vs_hde.figure.display_selection()
            self.es_hde.figure.display_selection()
        elif tab == 2:
            self.tab2.update_region_table()
            self.update_region_callback(self, self.region_set)
        elif tab == 3:
            if len(self.tab2.selected_regions) == 0:
                self.select_tab(2)
            else:
                region = self.region_set.get(self.tab2.selected_regions[0]['Region'])
                self.tab3.update_region(region, False)
                if region is None:
                    region = ModelRegion(self.X, self.y, self.X_test, self.y_test, self.model, score=self.score)
                self.vs_hde.figure.display_region(region)
                self.es_hde.figure.display_region(region)
        if not front:
            self.widget.children[4].v_model = max(tab - 1, 0)
        self.vs_hde.set_tab(tab)
        self.es_hde.set_tab(tab)
        self.tab = tab
        self.disable_hde()

    # ==================== TAB 1 ==================== #

    def new_rule_selected_callback(self, selection_mask, rules_mask):
        self.select_tab(1)
        self.vs_hde.figure.display_rules(selection_mask, rules_mask)
        self.es_hde.figure.display_rules(selection_mask, rules_mask)

    def validate_rules_callback(self, region: Region):
        self.selection_changed(None, boolean_mask(self.X, True))
        region.validate()
        self.region_set.add(region)
        self.select_tab(2)

    # ==================== TAB 2 ==================== #

    def edit_region_callback(self, caller, region):
        self.tab1.update_region(region)
        self.selection_mask = region.mask
        self.select_tab(1)

    def update_region_callback(self, caller, region_set):
        self.vs_hde.figure.display_regionset(region_set)
        self.es_hde.figure.display_regionset(region_set)

    def substitute_model_callback(self, caller, region):
        self.select_tab(3)
        self.tab3.update_region(region)

    # ==================== TAB 3 ==================== #

    @log_errors
    def model_validation_callback(self, *args):
        self.select_tab(2)
        self.tab2.selected_regions = []



import ipyvuetify as v
import pandas as pd
from antakia_core.compute.model_subtitution.model_class import MLModel
from antakia_core.data_handler.region import Region, ModelRegion
from plotly.graph_objects import FigureWidget, Bar

from pdpbox.pdp import PDPIsolate, PDPInteract

from antakia.utils.stats import log_errors


class ModelExplorer:
    """
    This is ModelExplorer docstring.
    """

    def __init__(self, X: pd.DataFrame):
        self._build_widget()
        self.model: MLModel | None = None
        self.region: ModelRegion | None = None
        self.X = X
        pass

    def _build_widget(self):
        self.feature_importance_tab = v.TabItem(  # Tab 1) feature importances # 43
            class_="mt-2",
            children=[]
        )
        self.pdp_feature_select = v.Select()
        self.pdp_figure = v.Container()
        self.widget = v.Tabs(
            v_model=0,  # default active tab
            children=[
                         v.Tab(children=["Feature Importance"]),
                         v.Tab(children=["Partial Dependency"]),
                     ]
                     +
                     [
                         self.feature_importance_tab,
                         v.TabItem(  # Tab 2) Partial dependence
                             children=[
                                 v.Col(
                                     children=[
                                         self.pdp_feature_select,
                                         self.pdp_figure
                                     ]
                                 )
                             ]
                         ),  # End of v.TabItem #2
                     ]
        )
        self.pdp_feature_select.on_event('change', self.display_pdp)

    def update_selected_model(self, model: MLModel, region: ModelRegion):
        self.model = model
        self.region = region
        self.update_feature_importances()
        self.update_pdp_tab()

    def update_feature_importances(self):
        if self.model is not None:
            feature_importances = self.model.feature_importances_.sort_values(ascending=True)
            fig = Bar(x=feature_importances, y=feature_importances.index, orientation='h')
            self.figure_fi = FigureWidget(data=[fig])
            self.figure_fi.update_layout(
                autosize=True,
                margin={
                    't': 0,
                    'b': 0,
                    'l': 0,
                    'r': 0
                },
            )
            self.figure_fi._config = self.figure_fi._config | {"displaylogo": False}

            self.feature_importance_tab.children = [self.figure_fi]
        else:
            self.feature_importance_tab.children = []

    def update_pdp_tab(self):
        if self.pdp_feature_select.v_model not in self.X.columns and self.model is not None:
            features = list(self.model.feature_importances_.sort_values(ascending=False).index)
            self.pdp_feature_select.items = features
            self.pdp_feature_select.v_model = features[0]
        self.display_pdp()

    @log_errors
    def display_pdp(self, *args):
        if self.model is not None:
            selected_feature = self.pdp_feature_select.v_model
            if self.X[self.region.mask][selected_feature].nunique() > 1:
                predict_func = self.model.__class__.predict
                figure = PDPIsolate(
                    df=self.X.copy(), feature=selected_feature, feature_name=selected_feature,
                    model=self.model, model_features=self.X.columns, pred_func=predict_func,
                    n_classes=0  # regression
                ).plot()[0]
                self.figure_pdp = FigureWidget(figure)
                self.figure_pdp.update_layout(
                    autosize=True, width=None, height=None,
                    margin={
                        't': 0,
                        'b': 0,
                        'l': 0,
                        'r': 0
                    },
                )
                self.figure_pdp._config = self.figure_pdp._config | {"displaylogo": False}

                self.pdp_figure.children = [self.figure_pdp]
            else:
                self.pdp_figure.children = ['only one feature value, no pdp to display']
        else:
            self.pdp_figure.children = []

    def reset(self):
        self.model = None
        self.region = None
        self.update_feature_importances()
        self.update_pdp_tab()



from traitlets import traitlets
import ipyvuetify as v


class SubModelTable(v.VuetifyTemplate):
    headers = traitlets.List([]).tag(sync=True, allow_null=True)
    items = traitlets.List([]).tag(sync=True, allow_null=True)
    selected = traitlets.List([]).tag(sync=True, allow_null=True)
    template = traitlets.Unicode('''
        <template>
            <v-data-table
                v-model="selected"
                :headers="headers"
                :items="items"
                item-key="Sub-model"
                show-select
                single-select
                :hide-default-footer="false"
                @item-selected="tableselect"
            >

            <template v-slot:item.delta="{ item }">
              <v-chip :color="item.delta_color" label size="small">
                    {{ item.delta }}
              </v-chip>
            </template>
            </v-data-table>
        </template>
        ''').tag(sync=True)  # type: ignore
    disable_sort = False

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.callback = None

    # @click:row="tableclick"
    # def vue_tableclick(self, data):
    #     raise ValueError(f"click event data = {data}")

    def set_callback(self, callback: callable):  # type: ignore
        self.callback = callback

    def vue_tableselect(self, data):
        self.callback(data)



from functools import partial

import ipyvuetify as v

from antakia.utils.stats import log_errors, stats_logger


class ColorSwitch:
    def __init__(self, y, y_pred, update_callback):
        self.y = y
        self.y_pred = y_pred
        self.update_callback = partial(update_callback, self)
        self._build_widget()

    def _build_widget(self):
        self.widget = v.BtnToggle(  # 11
            class_="mr-3",
            mandatory=True,
            v_model="Y",
            children=[
                v.Tooltip(  # 110
                    bottom=True,
                    v_slots=[
                        {
                            'name': 'activator',
                            'variable': 'tooltip',
                            'children':
                                v.Btn(  # 1100
                                    v_on='tooltip.on',
                                    icon=True,
                                    children=[
                                        v.Icon(children=["mdi-alpha-y-circle-outline"])
                                    ],
                                    value="y",
                                    v_model=True,
                                ),
                        }
                    ],
                    children=['Display target values']
                ),
                v.Tooltip(  # 111
                    bottom=True,
                    v_slots=[
                        {
                            'name': 'activator',
                            'variable': 'tooltip',
                            'children':
                                v.Btn(  # 1110
                                    v_on='tooltip.on',
                                    icon=True,
                                    children=[v.Icon(children=["mdi-alpha-y-circle"])],
                                    value="y^",
                                    v_model=True,
                                ),
                        }
                    ],
                    children=['Display predicted values']
                ),
                v.Tooltip(  # 112
                    bottom=True,
                    v_slots=[
                        {
                            'name': 'activator',
                            'variable': 'tooltip',
                            'children':
                                v.Btn(  # 1120
                                    v_on='tooltip.on',
                                    icon=True,
                                    children=[v.Icon(children=["mdi-delta"])],
                                    value="residual",
                                    v_model=True,
                                ),
                        }
                    ],
                    children=['Display residual values']
                ),
            ],
        )

        self.widget.on_event("change", self.switch_color)

    @log_errors
    def switch_color(self, widget, event, data):
        """
        Called with the user clicks on the colorChoiceBtnToggle
        Allows change the color of the dots
        """

        # Color : a pd.Series with one color value par row
        color = None
        stats_logger.log('color_changed', {'color': data})
        if data == "y":
            color = self.y
        elif data == "y^":
            color = self.y_pred
        elif data == "residual":
            color = self.y - self.y_pred
        self.update_callback(color)